from jumpscale.loader import j
import sys

DEFAULT_APP_NAME = "init"


class Application:
    def __init__(self):
        self.appname = DEFAULT_APP_NAME
        sys.excepthook = j.tools.errorhandler.excepthook

    @property
    def process_id(self):
        return j.sals.process.get_my_process().pid

    def start(self, appname):
        self.appname = appname
        j.logger.set_appname(appname)

        if j.core.db:
            j.core.db.sadd("applications", self.appname)

        j.logger.info("Application {} is started, process id: {}", self.appname, self.process_id)

    def stop(self):
        j.logger.info("Application {} is stopped", self.appname)

        if j.core.db:
            j.core.db.srem("applications", self.appname)



"""
This module exposes a j object, which contains a reference to all sub namespaces and modules available under jumpscale

It creates a container classes/types dynamically, inject attributes into these classes with references to sub namespaces and modules,
and at the end, create an instance from this container class.
"""
import importlib
import os
import types

import jumpscale


def get_container_type(full_name: str) -> type:
    """
    get a new type to be used as a container

    Args:
        full_name (str): full name to distinguish from other types.

    Returns:
        type: a new type

    """
    cls_name = "".join([name.capitalize() for name in full_name.split(".")])
    return type(cls_name, (object,), {"__fullname": full_name,})


def get_lazy_import_property(name, root_module, container_type):
    def getter(self):
        inner_name = f"__{name}"
        if hasattr(self, inner_name):
            return getattr(self, inner_name)

        full_name = f"{root_module.__name__}.{name}"
        mod = importlib.import_module(full_name)
        if mod.__spec__.origin in ("namespace", None):
            # if this module is a namespace, create a new container type
            sub_container_type = get_container_type(full_name)
            # expose all sub-modules of this imported module under this new type too
            expose_all(mod, sub_container_type)
            new_module = sub_container_type()
        else:
            # if it's just a module
            if hasattr(mod, "export_module_as"):
                new_module = mod.export_module_as()
            else:
                new_module = mod

        setattr(self, inner_name, new_module)
        return new_module

    return property(getter)


def expose_all(root_module: types.ModuleType, container_type: type):
    """
    exposes all sub-modules and namespaces to be available under given container type (class)

    Args:
        root_module (types.ModuleType): module
        ns_type (type): namepace type (class)
    """

    for path in root_module.__path__:
        for name in os.listdir(path):
            if not os.path.isdir(os.path.join(path, name)) or name == "__pycache__":
                continue

            lazy_import_property = get_lazy_import_property(name, root_module, container_type)
            setattr(container_type, name, lazy_import_property)


class J:
    @property
    def logger(self):
        return self.core.logging

    @property
    def application(self):
        return self.core.application

    @property
    def config(self):
        return self.core.config

    @property
    def exceptions(self):
        return self.core.exceptions


expose_all(jumpscale, J)
j = J()



import inspect
import sys
import traceback

from jumpscale.loader import j


class ErrorHandler:
    def __init__(self):
        self.handlers = []

    def _format_lines(self, lines):
        return "".join(lines).strip()

    def _construct_stacktrace(self, frame):
        stacktrace = []
        while frame:
            frame_info = inspect.getframeinfo(frame)

            if frame_info.code_context:
                code_context = frame_info.code_context[0].strip()
            else:
                code_context = ""

            stacktrace.append(
                {
                    "filename": j.sals.fs.basename(frame_info.filename),
                    "filepath": frame_info.filename,
                    "context": frame_info.function,
                    "linenr": frame_info.lineno,
                    "code": code_context,
                }
            )
            frame = frame.tb_next
        return stacktrace

    def _handle_exception(self, ttype, tvalue, tb, level=40, die=False, log=True, category="", data=None):
        timestamp = j.data.time.now().timestamp
        stacktrace = self._construct_stacktrace(tb)
        message = self._format_lines(traceback.format_exception_only(ttype, tvalue))
        traceback_text = self._format_lines(traceback.format_exception(ttype, tvalue, tb))

        err_dict = {
            "appname": j.application.appname,
            "level": level,
            "message": message,
            "timestamp": timestamp,
            "category": category or "exception",
            "data": data,
            "traceback": {"raw": traceback_text, "stacktrace": stacktrace, "process_id": j.application.process_id},
        }

        if log:
            exception = (ttype, tvalue, tb)
            j.logger.exception(message=message, category=category, data=data, level=level, exception=exception)

        for handler_func, handler_level in self.handlers:
            if level >= handler_level:
                handler_func(**err_dict)

        if die:
            sys.exit(1)

    def handle_exception(
        self,
        exception: Exception,
        level: int = 40,
        die: bool = False,
        log: bool = True,
        category: str = "",
        data: dict = None,
    ):
        """Hndler exception

        Arguments:
            exception {Exception} -- the exception object to handle

        Keyword Arguments:
            level {int} -- exception level (default: {40})
            die {bool} -- optional flag to exit after handling the exception (default: {True})
            log {bool} -- option flag to log the exception (default: {True})
            category {str} -- category (default: {""})
        """
        ttype, _, tb = sys.exc_info()
        self._handle_exception(ttype, exception, tb, die=die, log=log, category=category, data=data)

    def register_handler(self, handler: callable, level: int = 40):
        """Register new error handler

        Arguments:
            handler {callable} -- error handler callable

        Keyword Arguments:
            level {int} -- exception level (default: {40})
        """
        if (handler, level) not in self.handlers:
            self.handlers.append((handler, level))

    def excepthook(self, ttype, tvalue, tb):
        """exception hook handler"""
        self._handle_exception(ttype, tvalue, tb)



import os, pwd, grp

__all__ = ["get_user_pwd", "get_current_pwd", "get_group_grp", "get_current_grp"]


def get_user_pwd(uid):
    """get user passwd record
    Args:
        uid (int): uid of the user

    Returns:
        pwd.struct_passwd
    """
    return pwd.getpwuid(uid)


def get_current_pwd():
    """get current user passwd record

    Returns:
        pwd.struct_passwd
    """
    return get_user_pwd(os.getuid())


def get_group_grp(gid):
    """get group info
    Args:
        gid (int): gid of the group

    Returns:
        grp.struct_group
    """
    return grp.getgrgid(gid)


def get_current_grp():
    """get current group info

    Returns:
        grp.struct_group
    """
    return get_group_grp(os.getgid())



"""
Redis client
"""

from redis import Redis, exceptions

from jumpscale.core import events

from jumpscale.clients.base import Client
from jumpscale.core.base import fields
from jumpscale.core.base.events import AttributeUpdateEvent


class RedisClientAttributeUpdated(AttributeUpdateEvent):
    pass


class RedisClient(Client):
    hostname = fields.String(default="localhost")
    port = fields.Integer(default=6379)
    password = fields.Secret(default="")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__client = None

    def _attr_updated(self, name, value):
        super()._attr_updated(name, value)
        # this will allow other people to listen to this event too
        event = RedisClientAttributeUpdated(self, name, value)
        events.notify(event)

        # reset client
        self.__client = None

    def __dir__(self):
        return list(self._fields.keys()) + dir(self.redis_client)

    @property
    def redis_client(self):
        if not self.__client:
            if self.password:
                self.__client = Redis(self.hostname, self.port, password=self.password)
            else:
                self.__client = Redis(self.hostname, self.port)

        return self.__client
    
    def is_running(self):
        try:
            return self.redis_client.ping()
        except exceptions.ConnectionError:
            return False

    def __getattr__(self, k):
        # forward non found attrs to self.redis_client
        return getattr(self.redis_client, k)



import better_exceptions
import pudb
import sys
import traceback

from functools import partial
from prompt_toolkit.application import get_app
from prompt_toolkit.keys import Keys
from prompt_toolkit.formatted_text import ANSI
from prompt_toolkit.shortcuts import print_formatted_text
from prompt_toolkit.completion import Completion

from ptpython.prompt_style import PromptStyle
from ptpython.utils import get_jedi_script_from_document


def patched_handle_exception(self, e):
    """
    a new handler for ptpython repl exceptions
    it will call excepthook after ommitting all this framework's calls from traceback

    for the original, see ptpython.repl.PythonInput._handle_exception
    """
    output = self.app.output

    t, v, tb = sys.exc_info()

    # Required for pdb.post_mortem() to work.
    sys.last_type, sys.last_value, sys.last_traceback = t, v, tb

    # loop until getting actual traceback
    last_stdin_tb = tb
    while tb:
        if tb.tb_frame.f_code.co_filename == "<stdin>":
            last_stdin_tb = tb
        tb = tb.tb_next

    # except hook does not work as expected
    sys.excepthook(t, v, last_stdin_tb)
    # just print formatted exception for now
    formatted = better_exceptions.format_exception(t, v, last_stdin_tb)
    print_formatted_text(ANSI(formatted))

    output.write("%s\n" % e)
    output.flush()


def sort_completions_key(completion):
    """
    sort completions according to their type

    Args:
        completion (jedi.api.classes.Completion): completion

    Returns:
        int: sorting order
    """
    if completion.type == "function":
        return 2
    elif completion.type == "instance":
        return 1
    else:
        return 3


def get_style_for_completion(completion):
    base = "bg:%s fg:ansiblack"
    if completion.type == "function":
        return base % "ansigreen"
    elif completion.type == "instance":
        return base % "ansiyellow"
    else:
        return base % "ansigray"


HIDDEN_PREFIXES = ("_", "__")


def get_completions(self, document, complete_event):
    """
    get completions filtered and colored

    To filter and color completions on type, we try get jedi completions first
    and check their type, because `prompt-toolkit.completion.Completion` does not contain type information
    """
    try:
        script = get_jedi_script_from_document(document, self.get_globals(), self.get_locals())
    except:
        return

    if script:
        try:
            reference = script.get_references()[0]
        except Exception:
            reference = ""

        for c in sorted(script.completions(), key=sort_completions_key):
            if c.name.startswith(HIDDEN_PREFIXES):
                if not reference or not reference.description.startswith("_"):
                    continue

            yield Completion(
                c.name_with_symbols,
                len(c.complete) - len(c.name_with_symbols),
                display=c.name_with_symbols,
                selected_style="bg:ansidarkgray",
                style=get_style_for_completion(c),
            )


def ptconfig(repl):
    repl.exit_message = "Bye!"
    repl.show_docstring = True

    # When CompletionVisualisation.POP_UP has been chosen, use this
    # scroll_offset in the completion menu.
    repl.completion_menu_scroll_offset = 0

    # Show line numbers (when the input contains multiple lines.)
    repl.show_line_numbers = True

    # Show status bar.
    repl.show_status_bar = True

    # When the sidebar is visible, also show the help text.
    # repl.show_sidebar_help = True

    # Highlight matching parethesis.
    repl.highlight_matching_parenthesis = True

    # Line wrapping. (Instead of horizontal scrolling.)
    repl.wrap_lines = True

    # Mouse support.
    repl.enable_mouse_support = False

    # Complete while typing. (Don't require tab before the
    # completion menu is shown.)
    # repl.complete_while_typing = True

    # Vi mode.
    repl.vi_mode = False

    # Paste mode. (When True, don't insert whitespace after new line.)
    repl.paste_mode = False

    # Use the classic prompt. (Display '>>>' instead of 'In [1]'.)
    repl.prompt_style = "classic"  # 'classic' or 'ipython'

    # Don't insert a blank line after the output.
    repl.insert_blank_line_after_output = False

    # History Search.
    # When True, going back in history will filter the history on the records
    # starting with the current input. (Like readline.)
    # Note: When enable, please disable the `complete_while_typing` option.
    #       otherwise, when there is a completion available, the arrows will
    #       browse through the available completions instead of the history.
    # repl.enable_history_search = False

    # Enable auto suggestions. (Pressing right arrow will complete the input,
    # based on the history.)
    repl.enable_auto_suggest = True

    # Enable open-in-editor. Pressing C-X C-E in emacs mode or 'v' in
    # Vi navigation mode will open the input in the current editor.
    # repl.enable_open_in_editor = True

    # Enable system prompt. Pressing meta-! will display the system prompt.
    # Also enables Control-Z suspend.
    repl.enable_system_bindings = False

    # Ask for confirmation on exit.
    repl.confirm_exit = False

    # Enable input validation. (Don't try to execute when the input contains
    # syntax errors.)
    # repl.enable_input_validation = True

    # Use this colorscheme for the code.
    repl.use_code_colorscheme("perldoc")

    # Set color depth (keep in mind that not all terminals support true color).
    repl.color_depth = "DEPTH_24_BIT"  # True color.

    repl.enable_syntax_highlighting = True

    repl.min_brightness = 0.3

    # Add custom key binding for PDB.

    @repl.add_key_binding(Keys.ControlB)
    def _debug_event(event):
        ' Pressing Control-B will insert "pdb.set_trace()" '
        event.cli.current_buffer.insert_text("\nimport pdb; pdb.set_trace()\n")

    @repl.add_key_binding(Keys.ControlJ)
    def _debug_event(event):
        """
        custom binding for pudb, to allow debugging a statement and also
        post-mortem debugging in case of any exception
        """
        import time
        b = event.cli.current_buffer
        app = get_app()
        statements = b.document.text
        if statements:
            import linecache

            linecache.cache["<string>"] = (len(statements), time.time(), statements.split("\n"), "<string>")
            app.exit(pudb.runstatement(statements))
            app.pre_run_callables.append(b.reset)
        else:
            pudb.pm()

    # Custom key binding for some simple autocorrection while typing.

    corrections = {"impotr": "import", "pritn": "print", "pr": "print("}

    @repl.add_key_binding(" ")
    def _(event):
        " When a space is pressed. Check & correct word before cursor. "
        b = event.cli.current_buffer
        w = b.document.get_word_before_cursor()
        if w is not None:
            if w in corrections:
                b.delete_before_cursor(count=len(w))
                b.insert_text(corrections[w])
        b.insert_text(" ")

    class CustomPrompt(PromptStyle):
        """
        The classic Python prompt.
        """

        def in_prompt(self):
            return [("class:prompt", "JS-NG> ")]

        def in2_prompt(self, width):
            return [("class:prompt.dots", "...")]

        def out_prompt(self):
            return []

    repl.all_prompt_styles["custom"] = CustomPrompt()
    repl.prompt_style = "custom"

    repl._handle_exception = partial(patched_handle_exception, repl)
    better_exceptions.hook()

    old_get_completions = repl._completer.completer.__class__.get_completions

    def custom_get_completions(self, document, complete_event):
        completions = []

        try:
            completions = list(get_completions(self, document, complete_event))

            if not completions:
                completions = old_get_completions(self, document, complete_event)
        except Exception:
            pass

        yield from completions

    repl._completer.completer.__class__.get_completions = custom_get_completions



def main():
    import os
    from jumpscale.loader import j

    with open("test_index.db", "w"):
        pass

    print(os.system("redis-cli --scan  --pattern 'test.indexer.quote.id://*' | xargs redis-cli del"))
    print(os.system("redis-cli --scan  --pattern 'test.quote://*' | xargs redis-cli del"))
    print(os.system("redis-cli --scan  --pattern 'test.quote.las*' | xargs redis-cli del"))
    print(os.system("redis-cli --scan  --pattern 'test.indexer.employee.id://*' | xargs redis-cli del"))
    print(os.system("redis-cli --scan  --pattern 'test.employee://*' | xargs redis-cli del"))
    print(os.system("redis-cli --scan  --pattern 'test.employee.las*' | xargs redis-cli del"))
    print(os.system("redis-cli --scan  --pattern 'test.indexer.db.id://*' | xargs redis-cli del"))
    print(os.system("redis-cli --scan  --pattern 'test.db://*' | xargs redis-cli del"))
    print(os.system("redis-cli --scan  --pattern 'test.db.las*' | xargs redis-cli del"))
    j.clients.sonic.get("test").flush_collection("test")


if __name__ == "__main__":
    main()



"""
This module used to manage your ssh keys, get the public key, get the private key, generate key, write key to the file system,
delete key from the file system, load key from the file system.

# Using sshkey

```
ssh_cl = j.clients.sshkey.get("ssh_test")
```

### Load ssh key from file
```
ssh_cl.load_from_file_system()
```
### Generate ssh keys
```
ssh_cl.generate_keys()
```
### Get public key path
```
 ssh_cl.public_key_path  -> "/root/.config/jumpscale/sshkeys/tU59lc6P.pub"
```
### Writing ssh keys to files system
```
ssh_cl.write_to_filesystem()
```
"""
from jumpscale.clients.base import Client
from jumpscale.core.base import fields
from jumpscale.loader import j


class SSHKeyClient(Client):
    public_key = fields.String()
    private_key = fields.Secret()
    private_key_path = fields.Secret()
    passphrase = fields.Secret(default="")
    duration = fields.Integer()
    allow_agent = fields.Boolean()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.private_key_path and j.sals.fs.exists(self.private_key_path):
            self.load_from_file_system()

    def load_from_file_system(self):
        """ Load public key and private key from files using private key path and public key path
        e.g
            ssh_cl = j.clients.sshkey.get("ssh_test")
            ssh_cl.load_from_file_system()
        """
        self.public_key = j.sals.fs.read_file(self.public_key_path)
        self.private_key = j.sals.fs.read_file(self.private_key_path)

    def generate_keys(self):
        """Generate a new ssh key
        e.g
            ssh_cl = j.clients.sshkey.get("ssh_test")
            ssh_cl.generate_keys()
        """
        if not self.private_key_path:
            # TODO: make sure the new sshkey name doesn't exist.
            sshkeys_dir = j.sals.fs.join_paths(j.core.config.config_root, "sshkeys")
            j.sals.fs.mkdirs(sshkeys_dir)
            self.private_key_path = j.sals.fs.join_paths(sshkeys_dir, j.data.idgenerator.chars(8))
        if self.passphrase and len(self.passphrase) < 5:
            raise ValueError("invalid passphrase length: should be at least 5 chars.")
        cmd = 'ssh-keygen -f {} -N "{}"'.format(self.private_key_path, self.passphrase)
        rc, out, err = j.core.executors.run_local(cmd)
        if rc == 0:
            self.public_key = j.sals.fs.read_file(self.public_key_path)
            self.private_key = j.sals.fs.read_file(self.private_key_path)
        else:
            raise RuntimeError("couldn't create sshkey")

    @property
    def public_key_path(self):
        """ Get the public key path
        e.g
            ssh_cl = j.clients.sshkey.get("ssh_test")
            ssh_cl.public_key_path  -> "/root/.config/jumpscale/sshkeys/tU59lc6P.pub"
        Returns
            str: the path for public key

        """
        return "{}.pub".format(self.private_key_path)

    def write_to_filesystem(self):
        """ Write public key and private key to files using private key path and public key path.
        e.g
            ssh_cl = j.clients.sshkey.get("ssh_test")
            ssh_cl.write_to_filesystem()
        """
        if not self.private_key:
            raise RuntimeError("no private key to write")

        if not self.public_key:
            raise RuntimeError("no public key to write")

        j.sals.fs.write_file(self.private_key_path, self.private_key)
        j.sals.fs.write_file(self.public_key_path, self.public_key)

    def delete_from_filesystem(self):
        pass



"""
Store defines the interface for the backend storage, let it be filesystem or redis.

This module also defines the abstractions needed for any storage backend.

Every backend should be able to organize configuration for multiple instance given a location, also
read/write the config data in raw (string) format.
"""


import os

from abc import ABC, abstractmethod
from enum import Enum

from jumpscale.data.nacl import NACL
from jumpscale.data.serializers import base64
from jumpscale.core.config import Environment


# we will use this as a key field name to get it when searching
KEY_FIELD_NAME = "instance_name_"


class InvalidPrivateKey(Exception):
    """
    raised when the private key configured is invalid
    """


class StoreException(Exception):
    """
    raised by store backends
    """


class ConfigNotFound(Exception):
    """
    raised when a config is not found for an instance
    """


class Location:
    """
    dot-separated auto-location for any type

    for example, if we have a class in jumpscale/clients/redis/<type>
    location name will be jumpscale.clients.redis.<type>
    """

    def __init__(self, *name_list, type_=None):
        self.type = type_
        self.name_list = list(name_list)

    @property
    def name(self):
        """
        get dot seprated string from current name list

        Returns:
            str: dot separated string
        """
        return ".".join(self.name_list)

    @property
    def path(self):
        """
        get a filesystem path with from `name`, where dots are replaced by `os.sep`

        Returns:
            str: path
        """
        return os.path.join(*self.name.split("."))

    @classmethod
    def from_type(cls, type_):
        """
        get a location from any type/class

        Args:
            type_ (type): any type (class)

        Returns:
            Location: a location object
        """
        return cls(type_.__module__, type_.__name__)

    def __str__(self):
        args = "', '".join(self.name_list)
        cls_name = self.__class__.__name__
        return f"{cls_name}('{args}')"

    __repr__ = __str__

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)


class EncryptionMode(Enum):
    """
    An enum to select encryption mode based on loading or storing the data
    """

    Encrypt = 0
    Decrypt = 1


class EncryptionMixin:
    """
    A mixin that provides encrypt and decrypt methods, which can be used in any store
    """

    def encrypt(self, data):
        """encrypt data

        Args:
            data (str): input string

        Returns:
            bytes: encrypted data as byte string
        """
        if not isinstance(data, bytes):
            data = data.encode()
        return self.nacl.encrypt(data, self.public_key)

    def decrypt(self, data):
        """decrypt data

        Args:
            data (bytes): encrypted byte string

        Returns:
            str: decrypted data
        """
        return self.nacl.decrypt(data, self.public_key).decode()


class ConfigStore(ABC):
    """
    the interface every config store should implement:

    - `read(instance_name)`:  reads the data of this instance name
    - `write(instance_name, data)`: writes the data of this instance
    - `list_all(instance_name)`: lists all instance names
    - `delete(instance_name)`: delete instance data
    - `find(self, cursor_=None, limit_=None, **query)`: optional search method with query as field mapping
    """

    @abstractmethod
    def read(self, instance_name):
        pass

    @abstractmethod
    def write(self, instance_name, data):
        pass

    @abstractmethod
    def list_all(self):
        pass

    @abstractmethod
    def find(self, cursor_=None, limit_=None, **query):
        pass

    @abstractmethod
    def delete(self, instance_name):
        pass


class EncryptedConfigStore(ConfigStore, EncryptionMixin):
    """the base class for any config store backend"""

    def __init__(self, location, serializer):
        """
        the base for encrypted config store

        Args:
            location (Location)
            serializer (Serializer)

        Raises:
            InvalidPrivateKey: in case the private key is not configured
        """
        self.location = location
        self.serializer = serializer
        self.config_env = Environment()
        self.priv_key = base64.decode(self.config_env.get_private_key())
        self.nacl = NACL(private_key=self.priv_key)
        self.public_key = self.nacl.public_key.encode()

        if not self.priv_key:
            raise InvalidPrivateKey

    def _encrypt_value(self, value):
        """
        encrypt a single value

        Args:
            value (str): value

        Returns:
            str: decrypted value
        """
        return base64.encode(self.encrypt(value)).decode("ascii")

    def _decrypt_value(self, value):
        """
        decrypt a single value

        Args:
            value (str): value

        Returns:
            str: decrypted value
        """
        return self.decrypt(base64.decode(value))

    def _process_config(self, config, mode):
        """
        process current config according to encryption mode

        Args:
            config (dict): config dict (can be nested)
            mode (EncryptionMode)
        """
        new_config = {}
        for name, value in config.items():
            if name.startswith("__") and value is not None:
                if mode == EncryptionMode.Decrypt:
                    new_config[name.lstrip("__")] = self._decrypt_value(value)
                else:
                    # preserve __ to know it's an encrypted value
                    new_config[name] = self._encrypt_value(value)
            elif isinstance(value, dict):
                new_config[name] = self._process_config(value, mode)
            else:
                new_config[name] = value
        return new_config

    def get(self, instance_name):
        """
        get instance config

        Args:
            instance_name (str): instance name

        Returns:
            dict: instance config as dict
        """
        config = self.serializer.deserialize(self.read(instance_name))
        return self._process_config(config, EncryptionMode.Decrypt)

    def find(self, cursor_=None, limit_=None, **query):
        """
        a generic find, which do a linear search over all items

        if you want a better way, use a store which provides search

        Args:
            cursor_ (any, optional): an optional cursor, to start searching from. Defaults to None.
            limit_ (int, optional): results limit. Defaults to None.
            query: a mapping between field and value fo search by

        Returns:
            tuple: the new cursor, total result count and a generator for results
        """
        all_names = self.list_all()
        if not all_names:
            # empty result
            return cursor_, (name for name in all_names)

        all_count = len(all_names)
        if not limit_:
            limit_ = all_count

        start_search = not bool(cursor_)
        found = []

        for index in range(all_count):
            instance_name = all_names[index]
            if instance_name == cursor_:
                start_search = True

            if not start_search:
                continue

            data = self.get(instance_name)
            for name, value in query.items():
                if name in data:
                    data[KEY_FIELD_NAME] = instance_name
                    target_value = data[name]
                    if isinstance(target_value, str):
                        # just a simple normalization
                        value = str(value).lower()
                        target_value = target_value.lower()

                    if value == target_value and data not in found:
                        found.append(data)

            if len(found) >= limit_:
                break

        if index == all_count - 1:
            new_cursor = None
        else:
            new_cursor = all_names[index + 1]

        # return the new cursor, total found and a generator
        return new_cursor, len(found), (config for config in found)

    def save(self, instance_name, config):
        """
        save instance config

        Args:
            instance_name (str): name
            config (dict): config data, any key that starts with `__` will be encrypted

        Returns:
            bool: written or not
        """
        new_config = self._process_config(config, EncryptionMode.Encrypt)
        return self.write(instance_name, self.serializer.serialize(new_config))



"""helpers around string manipulation.


## Remove prefix 
```
JS-NG> j.data.text.removeprefix("ahhmed", "ah")                                                                                                                                          
'hmed'
```

## Remove suffix

```
JS-NG> j.data.text.removesuffix("ahhmed.3bot", ".3bot")                                                                                                                                  
'ahhmed'
```
"""

# TO BE REMOVED when https://www.python.org/dev/peps/pep-0616/ is implemented
def removeprefix(s: str, prefix: str) -> str:
    """Remove a prefix string `prefix` from a string `s`.

    Args:
        s (str): string the contains prefix we want to remove
        prefix (str): prefix we want to remove

    Returns:
        str: string without the prefix part

    """
    if s.startswith(prefix):
        return s[len(prefix):]
    else:
        return s[:]

def removesuffix(s: str, suffix: str) -> str:
    """Remove a suffix string `suffix` from a string `s`.

    Args:
        s (str): string the contains suffix we want to remove
        suffix (str): suffix we want to remove

    Returns:
        str: string without the suffix part

    """
    if suffix and s.endswith(suffix):
        return s[:-len(suffix)]
    else:
        return s[:]


from gevent import event

from jumpscale.core.base import Base


class ServerRack(Base):
    def __init__(self):
        super().__init__()
        self._servers = dict()
        self._started = set()
        self._running_event = event.Event()

    def _start(self, server_name: str):
        if server_name not in self._started:
            self._servers[server_name].start()
            self._started.add(server_name)

    def _stop(self, server_name: str):
        if server_name in self._started:
            self._servers[server_name].stop()
            self._started.remove(server_name)

    def add(self, server_name, server):
        """Add new server to the rack

        Args:
            name (str): server name
            server (object): server object
        """
        if server_name in self._servers:
            self._stop(server_name)

        self._servers[server_name] = server

    def remove(self, server_name: str):
        """Remove server (Stop it first if running)

        Args:
            server_name (str): server name
        """
        self._stop(server_name)

        if server_name in self._servers:
            self._servers.pop(server_name)

        if server_name in self._started:
            self._started.remove(server_name)

    def start(self, server_name: str = None, wait: bool = False):
        """Start server by its name or start all servers

        Args:
            server_name (str, optional): server name, if None will start all the servers. Defaults to None.
        """
        self._running_event.clear()
        if server_name:
            self._start(server_name)
        else:
            for server_name in self._servers:
                self._start(server_name)

        if wait:
            try:
                self._running_event.wait()
            except KeyboardInterrupt:
                self.stop()

    def stop(self, server_name: str = None):
        """Stop server by its name or stop all running servers

        Args:
            server_name (str, optional): server name, if None will stop all running servers. Defaults to None.
        """
        if server_name:
            self._stop(server_name)
        else:
            for server_name in self._servers:
                self._stop(server_name)
        self._running_event.set()

    def is_running(self, server_name):
        return server_name in self._started



"""This module helps around creation of terminal tables

JS-NG> j.data.terminaltable.print_table("users", [ ["id", "name"], ["1", "ahmed"], ["2", "xmonader"]])           
+users----------+
| id | name     |
+----+----------+
| 1  | ahmed    |
| 2  | xmonader |
+----+----------+

JS-NG> tbl = j.data.terminaltable.create("users", [ ["id", "name"], ["1", "ahmed"], ["2", "xmonader"]])          
JS-NG> print(tbl)                                                                                                
+users----------+
| id | name     |
+----+----------+
| 1  | ahmed    |
| 2  | xmonader |
+----+----------+
"""
import terminaltables as tt

TABLE_TYPES = {"ascii": tt.AsciiTable, "single": tt.SingleTable, "double": tt.DoubleTable}


def create(title, data, type_="ascii"):

    table_type = TABLE_TYPES.get(type_)
    if not table_type:
        raise ValueError("invalid type {} allowed types are {}".format(table_type, TABLE_TYPES.keys()))

    table = table_type(data, title)

    return table.table


def print_table(title, data, type_="ascii"):
    print(create(title, data, type_))



"""
Exceptions module provide a carefully picked list of exceptions to be used across the framework

- Base
- Permission
- Halt
- Runtime
- Input
- Value
- NotImplemented
- Bug
- Operations
- IO
- NotFound
- Timeout
- SSH
- SSHTimeout


"""


class JSException(Exception):
    def __init__(self, message, category=None, level=None, context=None):
        super().__init__(message)


class Base(JSException):
    pass


class Permission(JSException):
    pass


class Halt(JSException):
    pass


class Runtime(JSException):
    pass


class Input(JSException):
    pass


class Value(JSException):
    pass


class NotImplemented(JSException):
    pass


class Validation(JSException):
    pass


class Bug(JSException):
    pass


class Operations(JSException):
    pass


class IO(JSException):
    pass


class NotFound(JSException):
    pass


class Timeout(JSException):
    pass


class SSH(JSException):
    pass


class SSHTimeout(JSException):
    pass



"""This module helps with getting all of the information of the platform related to the machine hosting js-ng


## Get CPU count

```
JS-NG> j.data.platform.CPU_COUNT
4
```

## Check if 64-bit system

```
JS-NG> j.data.platform.IS_64BIT
True
```

## Check if linux or osx

```
JS-NG> j.data.platform.is_linux()
True

JS-NG> j.data.platform.is_osx()
False

JS-NG> j.data.platform.is_unix()
True

```

## Get linux info
```

JS-NG> j.data.platform.lsb_release_info()
{'distributor_id': 'Ubuntu', 'description': 'Ubuntu 18.04.4 LTS', 'release': '18.04', 'codename': 'bionic'}

JS-NG> j.data.platform.os_release_info()
{'name': 'Ubuntu', 'version': '18.04.4 LTS (Bionic Beaver)', 'id': 'ubuntu', 'id_like': 'debian', 'pretty_name': 'Ubuntu 18.04.4 LTS', 'version_id': '18.04', 'home_url': 'https://www.ubuntu.com/', 'support_url': 'https://help.ubuntu.com/', 'bug_report_url': 'https://bugs.launchpad.net/ubuntu/', 'privacy_policy_url': 'https://www.ubuntu.com/legal/terms-and-policies/privacy-policy', 'version_codename': 'bionic', 'ubuntu_codename': 'bionic', 'codename': 'bionic'}
JS-NG> j.data.platform.name
<function name at 0x7f267e1072f0>

JS-NG> j.data.platform.name()
'Ubuntu'


JS-NG> j.data.platform.main()
Name: Ubuntu 18.04.4 LTS
Version: 18.04 (bionic)
Codename: bionic
JS-NG> j.data.platform.version()
'18.04'
'Ubuntu 18.04.3 LTS', 'release': '18.04', 'codename': 'bionic'}

```

## Get python info

```
JS-NG> j.data.platform.get_python_info()
{'argv': 'jsng', 'bin': '/home/ahmed/.cache/pypoetry/virtualenvs/js-ng--jYQyIYz-py3.6/bin/python', 'version': '3.6.9 (default, Apr 18 2020, 01:56:04) [GCC 8.4.0]', 'compiler': 'GCC 8.4.0', 'build_date': 'Apr 18 2020 01:56:04', 'version_info': [3, 6, 9, 'final', 0], 'features': {'openssl': 'OpenSSL 1.1.1  11 Sep 2018', 'expat': 'expat_2.2.5', 'sqlite': '3.22.0', 'tkinter': '', 'zlib': '1.2.11', 'unicode_wide': True, 'readline': True, '64bit': True, 'ipv6': True, 'threading': True, 'urandom': True}}
```

## Get the complete profile info

```
JS-NG> j.data.platform.get_profile()
{'username': 'xmonader', 'guid': 'ce5a52d09920b248b602caf28539073', 'hostname': 'xmonader-ThinkPad-E580', 'hostfqdn': 'xmonader-Th
inkPad-E580', 'uname': {'system': 'Linux', 'node': 'xmonader-ThinkPad-E580', 'release': '4.15.0-55-generic', 'version': '#60-Ubunt
u SMP Tue Jul 2 18:22:20 UTC 2019', 'machine': 'x86_64', 'processor': 'x86_64'}, 'linux_dist_name': 'debian', 'linux_dist_version'
: 'buster/sid', 'cpu_count': 8, 'fs_encoding': 'utf-8', 'ulimit_soft': 1024, 'ulimit_hard': 1048576, 'cwd': '/home/xmonader/wspace
/threefoldtech/js-ng', 'umask': '0o2', 'python': {'argv': 'jsng', 'bin': '/home/xmonader/.cache/pypoetry/virtualenvs/js-ng-py3.7/bin/pyt
hon', 'version': '3.7.4 (default, Jul 12 2019, 20:57:46) [GCC 5.4.0 20160609]', 'compiler': 'GCC 5.4.0 20160609', 'build_date': 'J
ul 12 2019 20:57:46', 'version_info': [3, 7, 4, 'final', 0], 'features': {'openssl': 'OpenSSL 1.0.2s  28 May 2019', 'expat': 'expa
t_2.2.7', 'sqlite': '3.29.0', 'tkinter': '', 'zlib': '1.2.11', 'unicode_wide': True, 'readline': True, '64bit': True, 'ipv6': True
, 'threading': True, 'urandom': True}}, 'time_utc': '2019-08-18 11:24:23.636866', 'time_utc_offset': 2.0}

```
"""

# from boltons.ecoutils

import re
import os
import sys
import time
import pprint
import random
import socket
import struct
import getpass
import datetime
import platform

try:
    getrandbits = random.SystemRandom().getrandbits
    HAVE_URANDOM = True
except Exception:
    HAVE_URANDOM = False
    getrandbits = random.getrandbits


# 128-bit GUID just like a UUID, but backwards compatible to 2.4
INSTANCE_ID = hex(getrandbits(128))[2:-1].lower()

IS_64BIT = struct.calcsize("P") > 4
HAVE_UCS4 = getattr(sys, "maxunicode", 0) > 65536
HAVE_READLINE = True

try:
    import readline
except Exception:
    HAVE_READLINE = False

try:
    import sqlite3

    SQLITE_VERSION = sqlite3.sqlite_version
except Exception:
    # note: 2.5 and older have sqlite, but not sqlite3
    SQLITE_VERSION = ""


try:

    import ssl

    try:
        OPENSSL_VERSION = ssl.OPENSSL_VERSION
    except AttributeError:
        # This is a conservative estimate for Python <2.6
        # SSL module added in 2006, when 0.9.7 was standard
        OPENSSL_VERSION = "OpenSSL >0.8.0"
except Exception:
    OPENSSL_VERSION = ""


try:
    import Tkinter as tkinter

    TKINTER_VERSION = str(tkinter.TkVersion)
except Exception:
    TKINTER_VERSION = ""


try:
    import zlib

    ZLIB_VERSION = zlib.ZLIB_VERSION
except Exception:
    ZLIB_VERSION = ""


try:
    from xml.parsers import expat

    EXPAT_VERSION = expat.EXPAT_VERSION
except Exception:
    EXPAT_VERSION = ""


try:
    from multiprocessing import cpu_count

    CPU_COUNT = cpu_count()
except Exception:
    CPU_COUNT = 0

try:
    import threading

    HAVE_THREADING = True
except Exception:
    HAVE_THREADING = False


try:
    HAVE_IPV6 = socket.has_ipv6
except Exception:
    HAVE_IPV6 = False


try:
    from resource import getrlimit, RLIMIT_NOFILE

    RLIMIT_FDS_SOFT, RLIMIT_FDS_HARD = getrlimit(RLIMIT_NOFILE)
except Exception:
    RLIMIT_FDS_SOFT, RLIMIT_FDS_HARD = 0, 0


START_TIME_INFO = {"time_utc": str(datetime.datetime.utcnow()), "time_utc_offset": -time.timezone / 3600.0}


def get_python_info():
    ret = {}
    ret["argv"] = _escape_shell_args(sys.argv)
    ret["bin"] = sys.executable

    # Even though compiler/build_date are already here, they're
    # actually parsed from the version string. So, in the rare case of
    # the unparsable version string, we're still transmitting it.
    ret["version"] = " ".join(sys.version.split())

    ret["compiler"] = platform.python_compiler()
    ret["build_date"] = platform.python_build()[1]
    ret["version_info"] = list(sys.version_info)

    ret["features"] = {
        "openssl": OPENSSL_VERSION,
        "expat": EXPAT_VERSION,
        "sqlite": SQLITE_VERSION,
        "tkinter": TKINTER_VERSION,
        "zlib": ZLIB_VERSION,
        "unicode_wide": HAVE_UCS4,
        "readline": HAVE_READLINE,
        "64bit": IS_64BIT,
        "ipv6": HAVE_IPV6,
        "threading": HAVE_THREADING,
        "urandom": HAVE_URANDOM,
    }

    return ret


def get_profile(scrub=False):
    """The main entrypoint to platform. Calling this will return a
    JSON-serializable dictionary of information about the current
    process.
    It is very unlikely that the information returned will change
    during the lifetime of the process, and in most cases the majority
    of the information stays the same between runs as well.
    :func:`get_profile` takes one optional keyword argument, *scrub*,
    a :class:`bool` that, if True, blanks out identifiable
    information. This includes current working directory, hostname,
    Python executable path, command-line arguments, and
    username. Values are replaced with '-', but for compatibility keys
    remain in place.
    """
    ret = {}
    try:
        ret["username"] = getpass.getuser()
    except Exception:
        ret["username"] = ""
    ret["guid"] = str(INSTANCE_ID)
    ret["hostname"] = socket.gethostname()
    ret["hostfqdn"] = socket.getfqdn()
    uname = platform.uname()
    ret["uname"] = {
        "system": uname[0],
        "node": uname[1],
        "release": uname[2],  # linux: distro name
        "version": uname[3],  # linux: kernel version
        "machine": uname[4],
        "processor": uname[5],
    }
    try:
        linux_dist = platform.linux_distribution()
    except Exception:
        linux_dist = ("", "", "")
    ret["linux_dist_name"] = linux_dist[0]
    ret["linux_dist_version"] = linux_dist[1]
    ret["cpu_count"] = CPU_COUNT

    ret["fs_encoding"] = sys.getfilesystemencoding()
    ret["ulimit_soft"] = RLIMIT_FDS_SOFT
    ret["ulimit_hard"] = RLIMIT_FDS_HARD
    ret["cwd"] = os.getcwd()
    ret["umask"] = oct(os.umask(os.umask(2))).rjust(3, "0")

    ret["python"] = get_python_info()
    ret.update(START_TIME_INFO)

    if scrub:
        # mask identifiable information
        ret["cwd"] = "-"
        ret["hostname"] = "-"
        ret["hostfqdn"] = "-"
        ret["python"]["bin"] = "-"
        ret["python"]["argv"] = "-"
        ret["uname"]["node"] = "-"
        ret["username"] = "-"

    return ret


_real_safe_repr = pprint._safe_repr


def _fake_json_dumps(val, indent=2):
    # never do this. this is a hack for Python 2.4. Python 2.5 added
    # the json module for a reason.
    def _fake_safe_repr(*a, **kw):
        res, is_read, is_rec = _real_safe_repr(*a, **kw)
        if res == "None":
            res = "null"
        if res == "True":
            res = "true"
        if res == "False":
            res = "false"
        if not (res.startswith("'") or res.startswith("u'")):
            res = res
        else:
            if res.startswith("u"):
                res = res[1:]

            contents = res[1:-1]
            contents = contents.replace('"', "").replace(r"\"", "")
            res = '"' + contents + '"'
        return res, is_read, is_rec

    pprint._safe_repr = _fake_safe_repr
    try:
        ret = pprint.pformat(val, indent=indent)
    finally:
        pprint._safe_repr = _real_safe_repr

    return ret


def get_profile_json(indent=False):
    if indent:
        indent = 2
    else:
        indent = 0
    try:
        import json

        def dumps(val, indent):
            if indent:
                return json.dumps(val, sort_keys=True, indent=indent)
            return json.dumps(val, sort_keys=True)

    except ImportError:

        def dumps(val, indent):
            ret = _fake_json_dumps(val, indent=indent)
            if not indent:
                ret = re.sub("\n\s*", " ", ret)
            return ret

    data_dict = get_profile()
    return dumps(data_dict, indent)


def _escape_shell_args(args, sep=" ", style=None):
    if not style:
        if sys.platform == "win32":
            style = "cmd"
        else:
            style = "sh"

    if style == "sh":
        return _args2sh(args, sep=sep)
    elif style == "cmd":
        return _args2cmd(args, sep=sep)

    raise ValueError("style expected one of 'cmd' or 'sh', not %r" % style)


_find_sh_unsafe = re.compile(r"[^a-zA-Z0-9_@%+=:,./-]").search


def _args2sh(args, sep=" "):
    # see strutils
    ret_list = []

    for arg in args:
        if not arg:
            ret_list.append("''")
            continue
        if _find_sh_unsafe(arg) is None:
            ret_list.append(arg)
            continue
        # use single quotes, and put single quotes into double quotes
        # the string $'b is then quoted as '$'"'"'b'
        ret_list.append("'" + arg.replace("'", "'\"'\"'") + "'")

    return " ".join(ret_list)


def _args2cmd(args, sep=" "):
    # see strutils
    result = []
    needquote = False
    for arg in args:
        bs_buf = []

        # Add a space to separate this argument from the others
        if result:
            result.append(" ")

        needquote = (" " in arg) or ("\t" in arg) or not arg
        if needquote:
            result.append('"')

        for c in arg:
            if c == "\\":
                # Don't know if we need to double yet.
                bs_buf.append(c)
            elif c == '"':
                # Double backslashes.
                result.append("\\" * len(bs_buf) * 2)
                bs_buf = []
                result.append('\\"')
            else:
                # Normal char
                if bs_buf:
                    result.extend(bs_buf)
                    bs_buf = []
                result.append(c)

        # Add remaining backslashes, if any.
        if bs_buf:
            result.extend(bs_buf)

        if needquote:
            result.extend(bs_buf)
            result.append('"')

    return "".join(result)


def is_linux():
    return sys.platform.lower() == "linux"


def is_osx():
    return sys.platform.lower() == "darwin"


def is_unix():
    return is_linux() or is_osx()


if is_linux():
    from distro import *



import binascii

import requests
from collections import namedtuple

from jumpscale.data.encryption import mnemonic
from jumpscale.data.nacl.jsnacl import NACL
from jumpscale.data.serializers import base64
from jumpscale.core import exceptions
from jumpscale.tools.console import ask_choice, ask_string, printcolors

NETWORKS = {"mainnet": "explorer.grid.tf", "testnet": "explorer.testnet.grid.tf", "devnet": "explorer.devnet.grid.tf"}

RESTART_CHOICE = "Restart from the begining"
REENTER_CHOICE = "Re-Enter your value"
CHOICES = [RESTART_CHOICE, REENTER_CHOICE]

IdentityInfo = namedtuple("IdentityInfo", ["identity", "email", "words", "explorer"])


class IdentityManager:
    def __init__(self, identity: str = "", email: str = None, words: str = None, explorer: str = None):
        self.identity = identity
        self.email = email
        self.words = words
        self.explorer = explorer

    def reset(self):
        self.identity = ""
        self.email = ""
        self.words = ""
        self.explorer = ""

    def _check_keys(self, user_explorer_key, user_app):
        if not user_app:
            return True
        pub_key_app = base64.decode(user_app["publicKey"])
        if binascii.unhexlify(user_explorer_key) != pub_key_app:
            return False
        return True

    def _get_user(self):
        response = requests.get(f"https://login.threefold.me/api/users/{self.identity}")
        if response.status_code == 404:
            raise exceptions.Value(
                "\nThis identity does not exist in 3bot mobile app connect, Please create an idenity first using 3Bot Connect mobile Application\n"
            )
        userdata = response.json()

        resp = requests.get("https://{}/explorer/users".format(self.explorer), params={"name": self.identity})
        if resp.status_code == 404 or resp.json() == []:
            # creating new user
            user = {}
            user["name"] = userdata["doublename"]
            user["pubkey"] = base64.decode(userdata["publicKey"]).hex()
            printcolors(
                f"\nWelcome {{CYAN}}{userdata['doublename']}{{WHITE}}. Creating a new record on {{CYAN}}{self.explorer}{{RESET}}.\n"
            )
            return user, userdata
        else:
            users = resp.json()

            if not self._check_keys(users[0]["pubkey"], userdata):
                raise exceptions.Value(
                    f"\nYour 3bot on {self.explorer} seems to have been previously registered with a different public key.\n"
                    f"The identity of {self.identity} is mismatched with 3bot connect app"
                    "Please contact support.grid.tf to reset it.\n"
                    "Note: use the same email registered on the explorer to contact support otherwise we cannot reset the account.\n"
                )

            if users:
                return (users[0], userdata)
            return None, userdata

    def _check_email(self, email):
        resp = requests.get("https://{}/explorer/users".format(self.explorer), params={"email": email})
        users = resp.json()
        if users:
            if users[0]["name"] == self.identity:
                return True
            else:
                return False
        else:
            return True

    def ask_identity(self, identity=None, explorer=None):
        def _fill_identity_args(identity, explorer):
            def fill_words():
                words = ask_string("Copy the phrase from your 3bot Connect app here: ")
                self.words = words

            def fill_identity():
                identity = ask_string("what is your threebot name (identity)? ")
                if "." not in identity:
                    identity += ".3bot"
                self.identity = identity

            if identity:
                if self.identity != identity and self.identity:
                    self.reset()
                self.identity = identity

            if explorer:
                self.explorer = explorer
            elif not self.explorer:
                response = ask_choice(
                    "Which network would you like to register to? ", ["mainnet", "testnet", "devnet", "none"]
                )
                self.explorer = NETWORKS.get(response, None)
            if not self.explorer:
                return True

            user, user_app = None, None
            while not user:
                fill_identity()
                try:
                    user, user_app = self._get_user()
                except exceptions.Value as e:
                    response = ask_choice(f"{e}What would you like to do? ", CHOICES)
                    if response == RESTART_CHOICE:
                        return False

            while not self.email:
                self.email = ask_string("What is the email address associated with your identity? ")
                if self._check_email(self.email):
                    break
                else:
                    self.email = None
                    response = ask_choice(
                        "This email is currently associated with another identity. What would you like to do? ",
                        CHOICES,
                    )
                    if response == RESTART_CHOICE:
                        return False

            print("Configured email for this identity is {}".format(self.email))

            # time to do validation of words
            hexkey = None
            while True:
                if not self.words:
                    fill_words()
                try:
                    seed = mnemonic.mnemonic_to_key(self.words.strip())
                    hexkey = NACL(seed).get_verify_key_hex()
                    if (user and hexkey != user["pubkey"]) or not self._check_keys(hexkey, user_app):
                        raise Exception
                    else:
                        return True
                except Exception:
                    choice = ask_choice(
                        "\nSeems one or more more words entered is invalid.\nWhat would you like to do? ", CHOICES,
                    )
                    if choice == RESTART_CHOICE:
                        return False
                    fill_words()

        while True:
            if _fill_identity_args(identity, explorer):
                identity_info = IdentityInfo(self.identity, self.email, self.words, self.explorer)
                return identity_info



from jumpscale.loader import j
from jumpscale.core.base import Base, fields
from .utils import render_config_template


class Location(Base):
    name = fields.String()
    path_url = fields.String()
    is_auth = fields.Boolean(default=False)
    force_https = fields.Boolean(default=False)
    path_location = fields.String()
    index = fields.String()
    use_weblibs = fields.Boolean(default=False)
    ipaddr_dest = fields.String()
    port_dest = fields.Integer()
    path_dest = fields.String()
    connection_type = fields.String()
    location_type = fields.String()
    scheme = fields.String()
    config = fields.String()

    @property
    def path_cfg_dir(self):
        return f"{self.parent.path_cfg_dir}/{self.parent.instance_name}_locations"

    @property
    def path_cfg(self):
        return f"{self.path_cfg_dir}/{self.instance_name}.conf"

    @property
    def path_web(self):
        return self.parent.path_web

    def write_config(self, content=""):
        if not content:
            content = render_config_template(f"location_{self.location_type}", obj=self)
        j.sals.fs.write_file(self.path_cfg, content)

    def configure(self):
        """Config is a server config file of nginx (in text format)
        """
        j.sals.fs.mkdir(self.path_cfg_dir)

        if self.location_type in ["static", "spa"]:
            if not self.path_location.endswith("/"):
                self.path_location += "/"

        # if self.location_type == "proxy":  Uncomment when dependencies are handled
        #     j.sals.process.execute("moonc .", cwd=self.path_location)

        self.write_config(self.config)



"""Helps with timing functions and see how long they took

example
```
JS-NG> @j.tools.timer.timeit
       def fact(n):
           if n == 1:
               return 1
           else:
               return n * fact(n-1)
JS-NG> fact(1)
2020-04-09 10:41:57.175 | INFO     | jumpscale.tools.timer.timer:wrapper:11 - func fact with args: (1,), kwargs:
{} took 2.384185791015625e-06
1

JS-NG> fact(5)
2020-04-09 10:41:59.959 | INFO     | jumpscale.tools.timer.timer:wrapper:11 - func fact with args: (1,), kwargs:
{} took 1.1920928955078125e-06
2020-04-09 10:41:59.960 | INFO     | jumpscale.tools.timer.timer:wrapper:11 - func fact with args: (2,), kwargs:
{} took 0.0003504753112792969
2020-04-09 10:41:59.960 | INFO     | jumpscale.tools.timer.timer:wrapper:11 - func fact with args: (3,), kwargs:
{} took 0.0005877017974853516
2020-04-09 10:41:59.960 | INFO     | jumpscale.tools.timer.timer:wrapper:11 - func fact with args: (4,), kwargs:
{} took 0.0008087158203125
2020-04-09 10:41:59.960 | INFO     | jumpscale.tools.timer.timer:wrapper:11 - func fact with args: (5,), kwargs: {} took 0.0010216236114501953
120
```

"""
import time
from jumpscale.loader import j


def timeit(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        diff = end_time - start_time
        j.logger.info("func {} with args: {}, kwargs: {} took {}".format(func.__name__, args, kwargs, diff))
        return result

    return wrapper



from jumpscale.loader import j
from enum import Enum
from jumpscale.core.base import Base, fields
from .location import Location
from .utils import render_config_template, DIR_PATH


class Status(Enum):
    INIT = "init"
    INSTALLED = "installed"


class Website(Base):

    port = fields.Integer(default=80)
    ssl = fields.Boolean()
    domain = fields.String()
    path = fields.String()
    locations = fields.Factory(Location)

    @property
    def path_cfg_dir(self):
        return f"{self.parent.path_cfg_dir}/servers"

    @property
    def path_cfg(self):
        return f"{self.path_cfg_dir}/{self.instance_name}.http.conf"

    @property
    def path_web(self):
        return self.parent.path_web

    def configure(self):
        """Writes configuration of the website and its locations
        """

        j.sals.fs.mkdir(self.path_cfg_dir)
        config = render_config_template("website", base_dir=j.core.dirs.BASEDIR, website=self)
        j.sals.fs.write_file(self.path_cfg, config)

        for location_name in self.locations.list_all():
            location = self.locations.get(location_name)
            location.configure()


class OpenRestyServer(Base):
    status = fields.Enum(Status)
    websites = fields.Factory(Website)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._cmd = None
        self._path_web = None
        self._path_cfg_dir = None
        self._logs_dir = None

        self.executor = "tmux"  # only tmux for now

    @property
    def path_web(self):
        if not self._path_web:
            self._path_web = j.sals.fs.join_paths(j.core.dirs.VARDIR, "web", self.instance_name)
            j.sals.fs.mkdirs(j.sals.fs.join_paths(self._path_web, "static"))
        return self._path_web

    @property
    def path_cfg_dir(self):
        if not self._path_cfg_dir:
            self._path_cfg_dir = j.sals.fs.join_paths(j.core.dirs.CFGDIR, "nginx", self.instance_name)
            j.sals.fs.mkdirs(self._path_cfg_dir)
        return self._path_cfg_dir

    @property
    def path_cfg(self):
        return j.sals.fs.join_paths(self.path_cfg_dir, "nginx.conf")

    @property
    def logs_dir(self):
        if not self._logs_dir:
            self._logs_dir = j.sals.fs.join_paths(j.core.dirs.LOGDIR, "openresty", self.instance_name)
            j.sals.fs.mkdirs(self._logs_dir)
        return self._logs_dir

    def configure(self):
        # clean old websites config
        self.cleanup()
        """configures main nginx conf
        """
        # self.install() This is commented for now until the repo and necessary deps are handled
        configtext = j.tools.jinja2.render_template(
            template_path=j.sals.fs.join_paths(DIR_PATH, "templates", "nginx.conf"), logs_dir=self.logs_dir,
        )
        j.sals.fs.write_file(self.path_cfg, configtext)

    def get_from_port(self, port, domain=None, ssl=None):
        """will try to get a website listening on port, if it doesn't exist it will create one

        Args:
            port (int): port to search for
            domain (str, optional): domain. Defaults to None.
            ssl (bool, optional): Will set ssl if True. Defaults to None.

        Returns:
            Website: A new or an old website instance with the needed port
        """
        website_name = f"{self.instance_name}_website_{port}"

        website = self.websites.find(website_name)
        if website:
            return website

        website = self.websites.get(website_name)
        ssl = ssl or port == 443  # Use ssl if port is 443 if ssl in not specified

        website.domain = domain
        website.port = port
        website.ssl = ssl

        return website

    def install(self, reset=False):
        """Install required deps for openresty

        Args:
            reset (bool, optional): If true will redo the installation. Defaults to False.
        """
        if reset or self.status == "init":
            # get weblib
            weblibs_path = j.tools.git.ensure_repo(
                "https://github.com/threefoldtech/js-weblibs"  # Place holder repo might be changed
            )

            # copy the templates to the right location
            j.sals.fs.copy_tree(f"{DIR_PATH}/web_resources/", self.path_cfg_dir)

            j.sals.fs.symlink(
                f"{weblibs_path}/static", f"{self.path_web}/static/weblibs", overwrite=True,
            )
            self.status = Status.INSTALLED

            self.save()

    @property
    def startup_cmd(self):
        pass

    def start(self, reset=False):
        pass

    def stop(self):
        pass

    def is_running(self):
        pass

    def reload(self):
        self.configure()
        j.sals.process.execute("lapis build", cwd=self.path_cfg_dir)

    def cleanup(self):
        j.sals.fs.rmtree(f"{self.path_cfg_dir}/servers")



from jumpscale.loader import j


def _get_identifier(appname, message, public_message, category, alert_type):
    return j.data.hash.md5(":".join([appname, message, public_message, category, alert_type]))


class Alert:
    def __init__(self):
        self.id = None
        self.type = None
        self.level = 0
        self.appname = None
        self.category = None
        self.message = None
        self.public_message = None
        self.count = 0
        self.status = None
        self.first_occurrence = None
        self.last_occurrence = None
        self.data = None
        self.event = []
        self.tracebacks = []

    @classmethod
    def loads(cls, value):
        json = j.data.serializers.json.loads(value)
        instance = cls()
        instance.__dict__ = json
        return instance

    @property
    def identifier(self):
        return _get_identifier(self.appname, self.message, self.public_message, self.category, self.type)

    @property
    def json(self):
        return self.__dict__

    def dumps(self):
        return j.data.serializers.json.dumps(self.__dict__)


class AlertsHandler:
    def __init__(self):
        self._rkey = "alerts"
        self._rkey_id = "alerts:id"
        self._rkey_incr = "alerts:id:incr"
        self._db = None

    def __dir__(self):
        return ("get", "find", "alert_raise", "count", "reset", "delete", "delete_all")

    @property
    def db(self):
        if self._db is None:
            self._db = j.core.db
        return self._db

    def get(self, alert_id: int = None, identifier: str = None, die: bool = True) -> Alert:
        """Get alert by its id or identifier

        Keyword Arguments:
            alert_id {int} -- alert id (default: {None})
            identifier {str} -- alert identifier (default: {None})
            die {bool} -- flag to rasie exception if alert is not found (default: {True})

        Raises:
            j.core.exceptions.NotFound: alert is not found
            j.core.exceptions.Value: invalid arguments

        Returns:
            Alert -- [description]
        """
        if not (alert_id or identifier):
            raise j.core.exceptions.Value("Either alert id or alert identifier are required")

        alert_id = alert_id or self.db.hget(self._rkey_id, identifier)
        if alert_id:
            alert = self.db.hget(self._rkey, alert_id)
            if alert:
                return Alert.loads(alert)
        if die:
            raise j.core.exceptions.NotFound("Requested alert is not found")

    def find(
        self,
        appname: str = "",
        category: str = "",
        message: str = "",
        pid: int = None,
        start_time: int = None,
        end_time: int = None,
    ) -> Alert:

        """Find alerts

        Keyword Arguments:
            appname {str} -- filter by allert app name (default: {""})
            category {str} -- filter by alert category (default: {""})
            message {str} -- filter by alert message (default: {""})
            pid {int} -- filter by process id (default: {None})
            start_time {int} -- filter by start time (default: {None})
            end_time {int} -- filter by end time (default: {None})

        Returns:
            Alert -- alert object
        """
        appname = appname.strip().lower()
        category = category.strip().lower()
        message = message.strip().lower()

        alerts = []
        items = self.db.hscan_iter(self._rkey)

        for item in items:
            alert = Alert.loads(item[1])

            if appname and appname != alert.appname:
                continue

            if category and category != alert.category:
                continue

            if message and (message not in alert.message and message not in alert.public_message):
                continue

            if start_time and start_time < alert.first_occurrence:
                continue

            if end_time and end_time > alert.last_occurrence:
                continue

            if pid:
                for traceback in alert.tracebacks:
                    if traceback["process_id"] == pid:
                        break
                else:
                    continue

            alerts.append(alert)
        return sorted(alerts, key=lambda alert: alert.id)

    def alert_raise(
        self,
        appname,
        message,
        public_message: str = "",
        category: str = "",
        alert_type: str = "event_system",
        level: int = 40,
        data: dict = None,
        timestamp: float = None,
        traceback: dict = None,
    ) -> Alert:

        """Raise a new alert

        Arguments:
            message {str} -- alert message

        Keyword Arguments:
            public_message {str} -- alert public message (default: {""})
            category {str} -- alert category (default: {""})
            alert_type {str} -- alert type (default: {"event_system"})
            level {int} -- alert level (default: {40})
            traceback {dict} -- alert traceback (default: {None})

        Returns:
            Alert -- alert object
        """
        if not self.db.is_running():
            return

        identifier = _get_identifier(appname, message, public_message, category, alert_type)
        alert = self.get(identifier=identifier, die=False) or Alert()

        if alert.id:
            if alert.status == "new":
                alert.status = "open"
            elif alert.status == "closed":
                alert.status = "reopened"
        else:
            alert.status = "new"
            alert.first_occurrence = timestamp or j.data.time.now().timestamp

        alert.appname = appname
        alert.category = category
        alert.message = message
        alert.public_message = public_message
        alert.level = level
        alert.type = alert_type
        alert.count += 1
        alert.last_occurrence = timestamp or j.data.time.now().timestamp

        if traceback:
            if len(alert.tracebacks) > 5:
                alert.tracebacks.pop(0)

        alert.tracebacks.append(traceback)
        self._save(alert)
        return alert

    def count(self) -> int:
        """Gets alerts count

        Returns:
            int -- total number of alerts
        """
        return self.db.hlen(self._rkey)

    def _save(self, alert: Alert):
        """Saves alert object in db

        Arguments:
            alert {Alert} -- alert object
        """
        if not alert.id:
            alert.id = self.db.incr(self._rkey_incr)

        self.db.hset(self._rkey, alert.id, alert.dumps())
        self.db.hset(self._rkey_id, alert.identifier, alert.id)

    def delete(self, alert_id: int = None, identifier: str = None):
        """Delete alert by its id or identifier

        Raises:
            j.core.exceptions.Value: invalid arguments

        Keyword Arguments:
            alert_id {int} -- alert id (default: {None})
            identifier {str} -- alert identifier (default: {None})
        """
        if not (alert_id or identifier):
            raise j.core.exceptions.Value("Either alert id or alert identifier are required")

        alert_id = alert_id or self.db.hget(self._rkey_id, identifier)
        if alert_id:
            self.db.hdel(self._rkey, alert_id)

    def delete_all(self):
        """Deletes all alerts
        """
        self.db.delete(self._rkey, self._rkey_id)

    def reset(self):
        """Delete all alerts and reset the db
        """
        self.delete_all()
        self.db.delete(self._rkey_incr)



from jumpscale.loader import j
from jumpscale.data.schema import Property
from jumpscale.data.types import Integer, JSObject
import json


class JSObjBase:
    def __init__(self, model):
        self.model = model

    def get_dict(self):
        return self.model.get_dict(self)

    def save(self):
        return self.model.save_obj(self)

    def __str__(self):
        return json.dumps(self.get_dict(), indent=4)


class ModelBase:
    _schema = ""
    _name = ""

    def __init__(self, bcdb):
        self.schema = self._load_schema()
        self.schema.props["id"] = Property()
        self.schema.props["id"].unique = True
        self.schema.props["id"].index = True
        self.schema.props["id"].type = Integer()
        self.schema.props["id"].name = "id"

        self.bcdb = bcdb
        self.name = self._name

    def _load_schema(self):
        return j.data.schema.parse_schema(self._schema)

    def create_obj(self, data):
        """Create a new object and assign a new id to it with the given data dict.
        Missing props are defaultly initialized.

        Args:
            data (dict): The data dict that is used to create the object.

        Returns:
            JSObjBase: The newly created
        """
        o = JSObjBase(self)
        data["id"] = self._incr_id()
        self.set_from_dict(o, data)
        return o

    def _assert_uniqueness(self, obj):
        """Checks that the object doesn't contain an already existing property that is marked as unique.

        Args:
            obj (JSObjBase): The JS Object.

        Raises:
            RuntimeError: If it contains a duplicate of a unique value.
        """
        for prop in self.schema.props.values():
            if prop.unique:
                dbobj = self.get_by(prop.name, getattr(obj, prop.name))
                if dbobj is not None and dbobj.id != obj.id:
                    raise RuntimeError(f"{prop.name} is unique. One already exists.")

    def save_obj(self, obj):
        """Saves the object to the db. It forwards the call to the bcdb client.

        Args:
            obj (JSObjBase): The object to be saved.
        """
        self._assert_uniqueness(obj)
        self.bcdb.save_obj(self, obj)

    def _incr_id(self):
        """Increment the id counter and returns the newly incremented unique id.

        Returns:
            int: The new id.
        """
        return self.bcdb.model_id_incr(self)

    def get_by(self, key, value):
        """Search for objects whose key equal value.
        1. It searches in the redis index if key is indexed.
        2. Else, It's searched for in the sqlite index if the key is indexed for range search.
        3. Else, All objects in the db belonging to the given model is scanned linearly to determine the matching object.

        Args:
            key (str): The model property that is checked for.
            val (value): The value.

        Raises:
            RuntimeError: If the key is not a part of the schema.

        Returns:
            JSObjBase or None: The matched object (o: o.key == val). None if none matched.
        """
        return self.bcdb.get_entry(self, key, value)

    def get_range(self, key, min, max):
        return self.bcdb.get_item_from_index_set(self, key, min, max)

    def get_pattern(self, key, pattern):
        """Searches for objects whose key matches the given pattern in this model. The key must be registered in the text index.

        Args:
            key (str): The model property that the pattern is searched for in.
            pattern (str): The pattern to be searched for.

        Notes:
            Currently sonic server matches for some patterns and doesn't for others.

        Raises:
            RuntimeError: If the key is not defined in the model.
            RuntimeError: If the key is not indexed for search

        Returns:
            list[JSObjBase]: List of matching objects (o: o.key matches pattern).
        """
        return self.bcdb.get_item_from_index_text(self, key, pattern)

    def get_dict(self, obj):
        """Extracts a dict with all attributes from obj.

        Args:
            obj (JSObjBase): The object that the data is extracted from.

        Returns:
            dict: The data dict.
        """
        d = {}
        for prop in self.schema.props.values():
            prop_name = prop.name
            d[prop_name] = getattr(obj, prop_name)
            if isinstance(prop.type, JSObject):
                d[prop_name] = d[prop_name].get_dict()
        return d

    def set_from_dict(self, o, d):
        """Sets the attributes in the object o to the data from the dict d.
        Default values is used if it's not present in d.

        Args:
            o (JSObjBase): The object that will be set.
            d (dict): The dict containing the object new data.

        Notes:
            Type checking needs modification.
            It was assumed to receive strings convertable to the defined property.

        Raises:
            ValueError: If a value in the dict doesn't conform to the type checking rules imposed by the model's schema.

        Returns:
            JSObjBase: The JSObject after setting its data from the dict d.
        """
        for prop in self.schema.props.values():
            prop_name = prop.name
            if prop_name in d:
                if not prop.type.check(d[prop_name]):
                    raise ValueError("Wrong form")
                if isinstance(prop.type, JSObject):
                    obj_model = self.bcdb.get_model_by_name(prop.defaultvalue)
                    setattr(o, prop_name, obj_model.load_obj_from_dict(d[prop_name]))
                else:
                    setattr(o, prop_name, prop.type.from_str(d[prop_name]))
            else:
                setattr(o, prop_name, prop.type.default)
        return o

    def load_obj_from_dict(self, d):
        """Creates a new object with its data extracted from the d.

        Args:
            d (dict): The dict containing the newly created object's attributes.

        Returns:
            JSObjBase: The newly created object.
        """
        o = JSObjBase(self)
        self.set_from_dict(o, d)
        return o



import ast
import re


class String:
    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default

    def check(self, value):
        """Check whether provided value is a string"""
        return isinstance(value, str)

    def from_str(self, value):
        """return string from a string (is basically no more than a check)"""
        return value


class Integer:
    def __init__(self, default=None):
        if not default:
            default = 0
        self.default = self.from_str(default)

    def check(self, value):
        """Check whether provided string represent integer value

        Arguments:
            value (str)
        """
        try:
            int(value)
            return True
        except ValueError:
            return False

    def from_str(self, value):
        """get integer value from tha string

        Arguments:
            value (str)
        """
        try:
            return int(value)
        except ValueError:
            return None


class Boolean:
    def __init__(self, default=None):
        if not default:
            default = "False"
        self.default = self.from_str(default)

    def check(self, value):
        """Check whether provided string represent Boolean expresion

        Arguments:
            value (str)
        """
        valid_expresions = ["True", "False"]
        return value in valid_expresions

    def from_str(self, value):
        """Get the Boolean value from the string

        Arguments:
            value (str)
        """
        if value == "True":
            return True
        elif value == "False":
            return False
        else:
            return None


class Float:
    def __init__(self, default=None):
        if not default:
            default = 0.0
        self.default = self.from_str(default)

    def check(self, value):
        """Check whether provided string represent integer value

        Arguments:
            value (str)
        """
        try:
            float(value)
            return True
        except ValueError:
            return False

    def from_str(self, value):
        """get integer value from tha string

        Arguments:
            value (str)
        """
        try:
            return float(value)
        except ValueError:
            return None


class JSObject:
    def __init__(self, default=None):
        default = default or ""
        self.default = self.from_str(default)

    def check(self, value):
        return True

    def from_str(self, value):
        return value


class List:
    def __init__(self, default, subtype):
        self.subtype = subtype
        default = default or "[]"
        self.default = self.from_str(default)

    def _deep_check(self, value):
        """Check that the value represents a list with proper elements of the specified subtype.

        Args:
            value (list): The list to be checked.

        Returns:
            Boolean: True if the list is valid.
        """
        if not isinstance(value, list):
            return False
        for e in value:
            if isinstance(self.subtype, List) and not self.subtype._deep_check(e):
                return False
            elif not isinstance(self.subtype, List) and not self.subtype.check(str(e)):
                return False
        return True

    def check(self, value):
        """Check that the value represents a list with proper elements of the specified subtype.

        Args:
            value (list): The list to be checked.

        Returns:
            Boolean: True if the list is valid.
        """
        valid = len(value) >= 2 and value[0] == "[" and value[-1] == "]"
        if not valid:
            return False
        try:
            return self._deep_check(ast.literal_eval(value))
        except:
            return False
        return True

    def _deep_parse(self, value):
        """parses the subelements (if they are of different python type it's converted using the subtype parser)

        Args:
            value (list): The list to be parsed.

        Returns:
            list: The parsed list.
        """
        result = []
        if isinstance(self.subtype, List):
            for i, e in enumerate(value):
                value[i] = self.subtype._deep_parse(e)
        else:
            for i, e in enumerate(value):
                value[i] = self.subtype.from_str(str(e))
        return value

    def from_str(self, value):
        """parses the string value into a list.

        Args:
            value (str): The string to be parsed.

        Returns:
            list: The parsed list.
        """
        # print(f"from str-> value {value}")
        # print(f"subtype {self.subtype}")
        return object()
        # if isinstance(self.subtype, JSObject):
        #     print("returning noww")
        #     return object()
        # else:
        #     return self._deep_parse(ast.literal_eval(value))


def get_js_type(type_str, default_value=None):
    """Gets the js type from a string.

    1. "S" -> String
    2. "B" -> Boolean
    3. "I" -> Integer
    4. "O" -> JSObject
    4. "F" -> Float
    5. "L.*" -> List with subtype .*
    6. "E" -> String
    7. "" -> empty defaults to String

    Args:
        type_str (str): type description.
        default_value (any, optional): The default value. Defaults to None.

    Returns:
        Object: A js type object.
    """
    types = {
        "": String,
        "S": String,
        "I": Integer,
        "B": Boolean,
        "F": Float,
        "O": JSObject,
        "E": String,
        ## TODO: the following are totally wrong, but capturing them to be able to generate in python backend
        "email": String,
        "guid": String,
        "T": String,
        "D": String,
        "dict": String,
        "ipaddr": String,
        "ipaddress": String,
        "iprange": String,
        "bytes": String,
        "json": String,
        "I64": Float,
    }
    if len(type_str) == 0 or type_str[0] != "L":
        return types[type_str](default_value)
    else:
        subtype = type_str[1:]
        # print(subtype, default_value)
        return List(default_value, object)


class Email:
    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default
        self.regex = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"

    def check(self, value):
        """Check whether provided value is a valid email representation
        Arguments:
            value (str)
        Returbs:
            Boolean expresion"""
        return re.search(self.regex, value) is not None

    def from_str(self, value):
        """get the value from the string
        Arguments:
            value (str)
        Returbs:
            value (str)"""
        return value


class Path:
    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default
        self.regex = r"^(/[^/ ]*)+/?$"

    def check(self, value):
        """Check whether provided value is a valid path representation
        Arguments:
            value (str)
        Returbs:
            Boolean expresion"""
        return re.search(self.regex, value) is not None

    def from_str(self, value):
        """get the value from the string
        Arguments:
            value (str)
        Returbs:
            value (str)"""
        return value


class URL:
    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default
        self.regex = r"^(https?|ftp)://[^\s/$.?#].[^\s]*$"

    def check(self, value):
        """Check whether provided value is a valid URL representation
        Arguments:
            value (str)
        Returbs:
            Boolean expresion"""
        return re.search(self.regex, value) is not None

    def from_str(self, value):
        """get the value from the string
        Arguments:
            value (str)
        Returbs:
            value (str)"""
        return value


class Tel:
    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default
        self.regex = r"^\+?[0-9]{6,15}(?:x[0-9]+)?$"

    def check(self, value):
        """Check whether provided value is a valid Telephone number representation
        Arguments:
            value (str)
        Returbs:
            Boolean expresion"""
        value = self._clean(value)
        return re.search(self.regex, value) is not None

    def from_str(self, value):
        """get the value from the string
        Arguments:
            value (str)
        Returbs:
            value (str)"""
        return value

    def _clean(self, value):
        if value is not None:
            value = value.replace(",", "")
            value = value.replace("-", "")
            value = value.replace("(", "")
            value = value.replace(")", "")
            value = value.replace(" ", "")
            return value


class IPAddress:
    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default

    def check(self, value):
        """Check whether provided value is a valid IPaddress representation
        Arguments:
            value (str)
        Returns:
            Boolean expresion"""
        import ipaddress

        try:
            ipaddress.IPv4Address(value)
            return True
        except ipaddress.AddressValueError:
            return False

    def from_str(self, value):
        """get the value from the string
        Arguments:
            value (str)
        Returns:
            value (str)"""
        return value


class DateTime:
    """Supported format: yyyy-mm-dd hh:mm"""

    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default
        self.regex = r"^[0-9]{4}-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1]) (2[0-3]|[01][0-9]):[0-5][0-9]$"

    def check(self, value):
        """Check whether provided value is a valid datetime representation
        Arguments:
            value (str)
        Returbs:
            Boolean expresion"""
        return re.search(self.regex, value) is not None

    def from_str(self, value):
        """get the value from the string
        Arguments:
            value (str)
        Returbs:
            value (str)"""
        return value


class Date:
    """Support yyyy-mm-dd format"""

    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default
        self.regex = r"[0-9]{4}-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1])"

    def check(self, value):
        """Check whether provided value is a valid date representation
        Arguments:
            value (str)
        Returbs:
            Boolean expresion"""
        return re.search(self.regex, value) is not None

    def from_str(self, value):
        """get the value from the string
        Arguments:
            value (str)
        Returbs:
            value (str)"""
        return value


class Time:
    """Supported format : hh:mm"""

    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default
        self.regex = r"(2[0-3]|[01][0-9]):[0-5][0-9]"

    def check(self, value):
        """Check whether provided value is a valid time representation
        Arguments:
            value (str)
        Returbs:
            Boolean expresion"""
        return re.search(self.regex, value) is not None

    def from_str(self, value):
        """Get the value from string
            value (str)
        Returbs:
            value (str)"""
        return value


class Duration:
    """Supported format : (n)y (n)m (n)d (n)h (n)m (n)s"""

    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default
        self.regex = r"((\d{1,2}y\s?)?(\d{1,2}m\s?)?(\d{1,2}d\s?)?(\d{1,2}h\s?)?(\d{1,2}m\s?)?(\d{1,2}s\s?)?)|\d{1,2}"

    def check(self, value):
        """Check whether provided value is a valid duration representation
        Arguments:
            value (str)
        Returbs:
            Boolean expresion"""
        return re.search(self.regex, value) is not None

    def from_str(self, value):
        """Get the value from string
            value (str)
        Returbs:
            value (str)"""
        return value



from dataclasses import dataclass, field
from typing import List, Dict, Tuple, Callable
from enum import Enum


class NodeColor(Enum):
    White = 0
    Gray = 1
    Black = 2


def has_cycle_dfs(
    graph: Dict[str, List[str]],
    node: str,
    colors: Dict[str, NodeColor],
    has_cycle: List[bool],
    parent_map: Dict[str, str],
):
    if has_cycle[0]:
        return
    colors[node] = NodeColor.Gray

    for dep in graph[node]:
        parent_map[dep] = node
        if colors[dep] == NodeColor.Gray:
            has_cycle[0] = True
            parent_map["__CYCLESTART__"] = dep
            return
        if colors[dep] == NodeColor.White:
            has_cycle_dfs(graph, dep, colors, has_cycle, parent_map)
        colors[node] = NodeColor.Black


def graph_has_cycle(graph: Dict[str, List[str]]) -> Tuple[bool, Dict[str, str]]:
    colors: Dict[str, NodeColor] = {}
    for node in graph:
        colors[node] = NodeColor.White

    parent_map: Dict[str, str] = {}
    has_cycle = [False]
    for node in graph:
        parent_map[node] = "null"
        if colors[node] == NodeColor.White:
            has_cycle_dfs(graph, node, colors, has_cycle, parent_map)
        if has_cycle[0]:
            return (True, parent_map)
    return (False, parent_map)


@dataclass
class Task:
    name: str
    requires: List[str]
    action: Callable


@dataclass
class DepsResolver:
    tasksgraph: Dict[str, List[str]] = field(default_factory=dict)
    tasks: Dict[str, Task] = field(default_factory=dict)

    def add_task(self, task_name: str, deps: List[str], action: str):
        thetask = Task(task_name, deps, action)
        self.tasksgraph[task_name] = deps
        self.tasks[task_name] = thetask

    def run_task(self, task_name: str):
        has_cycle, parent_map = graph_has_cycle(self.tasksgraph)
        if has_cycle:
            print(f"cycle found: {parent_map}")
            return
        deps = []
        seen = []

        self._run_task_helper(task_name, deps, seen)
        for tsk in deps:
            t = self.tasks[tsk]
            if callable(t.action):
                t.action()
            elif isinstance(t.action, str):
                print("executing ", t.action)

    def _run_task_helper(self, task_name: str, deps: List[str], seen: List[str]):
        if task_name in seen:
            print(f"[+]resolved {task_name} before. no need to repeat action.")
        tsk = self.tasks[task_name]
        seen.append(task_name)
        if tsk.requires:
            for subtask in self.tasksgraph[tsk.name]:
                self._run_task_helper(subtask, deps, seen)
        deps.append(task_name)


if __name__ == "__main__":

    def with_cycle():
        deps_resolver = DepsResolver()
        deps_resolver.add_task("publish", ["build-release"], "print publish")
        deps_resolver.add_task("build-release", ["nim-installed"], "print exec command to build release mode")
        deps_resolver.add_task("nim-installed", ["curl-installed"], "print curl LINK | bash")
        deps_resolver.add_task("curl-installed", ["publish", "apt-installed"], "apt-get install curl")
        deps_resolver.add_task("apt-installed", [], "code to install apt...")
        deps_resolver.run_task("publish")

    def without_cycle():
        deps_resolver = DepsResolver()
        deps_resolver.add_task("publish", ["build-release"], "print publish")
        deps_resolver.add_task("build-release", ["nim-installed"], "print exec command to build release mode")
        deps_resolver.add_task("nim-installed", ["curl-installed"], "print curl LINK | bash")
        deps_resolver.add_task("curl-installed", ["apt-installed"], "apt-get install curl")
        deps_resolver.add_task("apt-installed", [], "code to install apt...")
        deps_resolver.run_task("publish")

    print("without cycle: ")
    without_cycle()

    print("with cycle")
    with_cycle()



from jumpscale.core.base import StoredFactory
from .client import ZDBClient, ZDBAdminClient


class ZDBFactory(StoredFactory):
    def new(self, name, *args, **kwargs):
        if kwargs.get("admin"):
            self.type = ZDBAdminClient
        return super().new(name, *args, **kwargs)


def export_module_as():

    return ZDBFactory(ZDBClient)



"""The compiler that parses JSX schema and generates a suitable backend in a supported lanaguage.

"""

from jumpscale.loader import j
import re

from .plugins import CrystalGenerator, JSNGGenerator


ALLOWED_LANGS = {"python": JSNGGenerator, "jsng": JSNGGenerator, "crystal": CrystalGenerator}


def generator_by_name(language_name="python"):
    """Gets a generator by name

    Keyword Arguments:
        language_name (str) -- suitable generator (default: {"python"})

    Returns:
        (jumpscale.tools.schema.SchemaGenerator.Plugin) -- The generator to use. Default
    """
    return ALLOWED_LANGS[language_name]


class Compiler:
    def __init__(self, lang="python", schema_text=""):
        """Compiler class responsible for parsing schema_text and creating Parsed Schema objects with all metadata information needed.

        Keyword Arguments:
            lang (str)-- language to generate for (default: {"python"})
            schema_text (str)-- the schema text.
        """
        self._schema_text = schema_text
        self.lang = lang = lang
        self._parsed_schemas = {}

    @property
    def generator(self):
        """Gets generator by `self.lang`."""
        return generator_by_name(self.lang)()

    def parse(self):
        """Parses all the schemas in `self._schema_text` and returns Schema objects for generation"""
        schemas_texts = []
        to_process = self._schema_text
        if to_process.count("@url") == 1:
            schemas_texts = [self._schema_text]
        else:
            urls_positions = [m.start() for m in re.finditer("@url", self._schema_text)]
            urls_positions.append(len(self._schema_text))
            start_end_positions = list(zip(urls_positions, urls_positions[1:]))

            schemas_texts = [self._schema_text[start:end] for (start, end) in start_end_positions]

        parsed_schemas = [j.data.schema.parse_schema(schema_text) for schema_text in schemas_texts]
        self._parsed_schemas = {s.url_to_class_name: s for s in parsed_schemas}

        return self._parsed_schemas



from jumpscale.loader import j


class Plugin:
    def _generate_single(self, schema):
        """Generates a single schema template

        Arguments:
            schema (jumpscale.data.Schema) -- Parsed schema.

        Returns:
            (str) -- schema convert to a class as string.
        """
        data = dict(
            generated_class_name=schema.url_to_class_name,
            generated_properties=schema.props,
            types_map=self._types_map,
            enums=schema.get_enums_required(),
            classes=schema.get_classes_required(),
            get_prop_line=self._get_prop_line,
        )
        return j.tools.jinja2.render_template(template_text=self._single_template, **data)

    def generate(self, parsed_schemas):
        """Generates a string has all the enumerations and classes in the target language.

        Arguments:
            parsed_schemas (Dict[str, jumpscale.data.Schema]) -- a dict of schema name to schema object

        Returns:
            (str) -- Compiled string of all enumerations and classes in the target language.
        """
        all_enums = []
        for scm_name, scm in parsed_schemas.items():
            enums = scm.get_enums_required()
            for enum in enums:
                if enum not in all_enums:
                    all_enums.append(enum)

        depsresolver = j.tools.depsresolver
        all_classes = []
        for scm_name, scm in parsed_schemas.items():
            depsresolver.add_task(scm.url_to_class_name, scm.get_classes_required(), lambda x: x)

        independent_schemas = [name for name, deps in depsresolver.tasksgraph.items() if not deps]
        for schema_name in independent_schemas:
            all_classes.append(self._generate_single(parsed_schemas[schema_name]))

        dependant_schemas = [name for name, deps in depsresolver.tasksgraph.items() if deps]
        for schema_name in dependant_schemas:
            all_classes.append(self._generate_single(parsed_schemas[schema_name]))

        all_classes_str = "\n\n".join(all_classes)
        return j.tools.jinja2.render_template(
            template_text=self._template, classes_generated=all_classes_str, enums=all_enums
        )



def export_module_as():

    from jumpscale.loader import j
    from redis.exceptions import ConnectionError

    class DefaultRedisDB:
        def __init__(self):
            self._db = None

        @property
        def db(self):
            if not self._db:
                self._db = j.clients.redis.get("main")
            return self._db

    return DefaultRedisDB().db



from redis import Redis
import json
from jumpscale.data.bcdb import models as models
from .clients import RedisStorageClient, RedisIndexClient, SonicIndexTextClient, SQLiteIndexSetClient

class BCDB:
    def __init__(self, ns):
        self.ns = ns
        self.storage = RedisStorageClient(ns)
        self.indexer = RedisIndexClient(ns)
        self.indexer_set = SQLiteIndexSetClient(ns)
        self.indexer_text = SonicIndexTextClient(ns)
        self.models = {
            
        }
        self.loaded_models = {

        }
        self.detect_models()
        self.model_model = self.models["model"](self)

    def detect_models(self):
        """It scans all the models and store its classes for later use."""
        for model_name in dir(models):
            model = getattr(models, model_name)
            if isinstance(model, type) and issubclass(model, models.ModelBase):
                self.models[model._name] = model
    
    def save_obj(self, model, obj):
        """Saves the given objects which belongs to model in the db and update the indexes.
        
        Args:
            model (ModelObj): The model object that obj belongs to.
            obj (JSObjBase): The object that will be saved.
        """
        self.indexer_set.set(model, obj)
        self.indexer_text.set(model, obj)
        for prop in model.schema.props.values():
            old_obj = model.get_by('id', obj.id)
            prop_name = prop.name
            index_prop = getattr(obj, prop.name)
            old_index = getattr(old_obj, prop.name) if old_obj else None
            if prop.index:
                self.indexer.set(model, prop_name, index_prop, obj.id, old_index)
        self.storage.set(model, obj.id, obj)

    def model_id_incr(self, model):
        """Increment the id counter in the model and returns the new id.
        Used to assign unique id for each created object.
        
        Args:
            model (ModelObj): The model object.
        
        Returns:
            int: The new unique id
        """
        return self.storage.incr_id(model)

    def get_item_by_id(self, model, id):
        """Gets the object in the model with the given id.
        
        Args:
            model (ModelObj): The model to be searched in.
            id (int): The object's id.
        
        Returns:
            JSObjBase or None: The JSObject with the given id. None if none was found.
        """
        return self.storage.get(model, id)

    def get_entry(self, model, key, val):
        """Search for objects whose key equal val.
        1. It searches in the redis index if key is indexed.
        2. Else, It's searched for in the sqlite index if the key is indexed for range search.
        3. Else, All objects in the db belonging to the given model is scanned linearly to determine the matching object.
        
        Args:
            model (ModelObj): The model in which the key is searched for.
            key (str): The model property that is checked for.
            val (value): The value.
        
        Raises:
            RuntimeError: If the key is not a part of the schema.
        
        Returns:
            JSObjBase or None: The matched object (o: o.key == val). None if none matched.
        """
        if key not in model.schema.props:
            raise RuntimeError(f"{key} is not a part of {model.name}'s schema")
        if model.schema.props[key].index:
            return self.get_item_from_index(model, key, val)
        elif model.schema.props[key].index_key:
            found = self.get_item_from_index_set(model, key, val, val)
            return found[0] if found else None
        else:
            for obj in self.storage.get_keys_in_model(model):
                if getattr(obj, key) == val:
                    return obj
        return None

    def get_range(self, model, key, min, max):
        """Searches for objects whose key lies between min and max.
        It tries to search for it in the index. If the key is not indexed it loops through all the objects.
        
        Args:
            model (ModelObj): The model in which the key is searched for.
            key (str): The model property that is checked for.
            min (value): The minimum.
            max (value): The maximum.
        
        Raises:
            RuntimeError: If the key is not a part of the schema.
        
        Returns:
            List[JSObjBase]: A list of matched objects (o: o.key >= min and o.key <= max)
        """
        if key not in model.schema.props:
            raise RuntimeError(f"{key} is not a part of {model.name}'s schema")
        if not model.schema.props[key].index_key:
            return self.get_item_from_index_set(model, key, min, max)
        else:
            result = []
            for obj in self.storage.get_keys_in_model(model):
                    obj_val = getattr(obj, key)
                    if obj_val >= min and obj_val <= max:
                        result.append(obj)
            return result
            

    def get_item_from_index(self, model, key, val):
        """Search for objects whose key equal val. The key must be indexed for search.
        
        Args:
            model (ModelObj): The model in which the key is searched for.
            key (str): The model property that is checked for.
            val (value): The value.
        
        Raises:
            RuntimeError: If the key is not a part of the schema.
            RuntimeError: If the key is not indexed for search.
        
        Returns:
            List[JSObjBase]: A list of matched objects (o: o.key == val)
        """
        if key not in model.schema.props:
            raise RuntimeError(f"{key} is not a part of {model.name}'s schema")
        if not model.schema.props[key].index:
            raise RuntimeError(f"{key} is not indexed.")
        keyid = self.indexer.get(model, key, val)
        return self.get_item_by_id(model, keyid) if keyid else None

    def get_item_from_index_set(self, model, key, min, max):
        """Searches for objects whose key lies between min and max. The key must be indexed for range search.
        
        Args:
            model (ModelObj): The model in which the key is searched for.
            key (str): The model property that is checked for.
            min (value): The minimum.
            max (value): The maximum.
        
        Raises:
            RuntimeError: If the key is not a part of the schema.
            RuntimeError: If the key is not indexed for range search.
        
        Returns:
            List[JSObjBase]: A list of matched objects (o: o.key >= min and o.key <= max)
        """
        if key not in model.schema.props:
            raise RuntimeError(f"{key} is not a part of {model.name}'s schema")
        if not model.schema.props[key].index_key:
            raise RuntimeError(f"{key} is not indexed.")
        return [self.get_item_by_id(model, x[0]) for x in self.indexer_set.get(model, key, min, max)]
        

    def get_item_from_index_text(self, model, key, pattern):
        """Searches for objects whose key matches the given pattern inside model. The key must be registered in the text index.
        
        Args:
            model (Modelobj): The model object in which the pattern is searched.
            key (str): The model property that the pattern is searched for in.
            pattern (str): The pattern to be searched for.
        
        Notes:
            Currently sonic server matches for some patterns and doesn't for others.

        Raises:
            RuntimeError: If the key is not defined in the model.
            RuntimeError: If the key is not indexed for search
        
        Returns:
            list[JSObjBase]: List of matching objects (o: o.key matches pattern).
        """
        if key not in model.schema.props:
            raise RuntimeError(f"{key} is not a part of {model.name}'s schema")
        if not model.schema.props[key].index_text:
            raise RuntimeError(f"{key} is not indexed for search.")
        return [self.get_item_by_id(model, int(x)) for x in self.indexer_text.get(model, key, pattern)]

    def get_model_by_name(self, model_name):
        """Returns a Model object given its name.
        
        Args:
            model_name (str): The name of the model.
        
        Raises:
            RuntimeError: Raised when no model exists with the given.
        
        Returns:
            ModelObj: The model object.
        """
        if model_name not in self.loaded_models:
            if model_name not in self.models:
                raise RuntimeError("Model not registered")
            self.loaded_models[model_name] = self.models[model_name](self)
        return self.loaded_models[model_name]



"""
This module defines main dirs in jumpscale to be used

```
JS-NG> j.core.dirs                                                                                                                                                        
ExportedModule(__doc__=None, _exportedas=<class 'jumpscale.core.dirs.dirs.Dirs'>, _loaded=True, _m=<module 'jumpscale.core.dirs' from '/home/ahmed/wspace/js/js-ng/jumpscale/core/dirs/__init__.py'>)

JS-NG> j.core.dirs.BASEDIR                                                                                                                                                
'/home/ahmed/sandbox'

JS-NG> j.core.dirs.BINDIR                                                                                                                                                 
'/home/ahmed/sandbox/bin'

JS-NG>                                                                                                                                                                    
JS-NG> j.core.dirs.CFGDIR                                                                                                                                                 
'/home/ahmed/sandbox/cfg'

JS-NG>                                                                                                                                                                    
JS-NG> j.core.dirs.CODEDIR                                                                                                                                                
'/home/ahmed/sandbox/code'

JS-NG>                                                                                                                                                                    
JS-NG> j.core.dirs.HOMEDIR                                                                                                                                                
'/home/ahmed'

JS-NG> j.core.dirs.LOGDIR                                                                                                                                                 
'/home/ahmed/sandbox/var/log'

JS-NG> j.core.dirs.TEMPLATEDIR                                                                                                                                            
'/home/ahmed/sandbox/var/templates'

JS-NG> j.core.dirs.TMPDIR                                                                                                                                                 
'/tmp/jumpscale'

JS-NG> j.core.dirs.VARDIR                                                                                                                                                 
'/home/ahmed/sandbox/var'
```


"""

import os


class Dirs:

    HOMEDIR = os.path.expanduser("~")  # TODO: ؤاثؤن homedir defined in sal.fs
    BASEDIR = os.path.join(HOMEDIR, "sandbox")
    BINDIR = os.path.join(BASEDIR, "bin")
    CFGDIR = os.path.join(BASEDIR, "cfg")  # TODO: check conflict with core.config_root..
    CODEDIR = os.path.join(BASEDIR, "code")
    VARDIR = os.path.join(BASEDIR, "var")
    LOGDIR = os.path.join(VARDIR, "log")
    TEMPLATEDIR = os.path.join(VARDIR, "templates")
    TMPDIR = "/tmp/jumpscale"



import inspect
import sys
from functools import wraps


def actor_method(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        # verify args and kwargs types
        signature = inspect.signature(func)
        bound = signature.bind(*args, **kwargs)
        for name, value in bound.arguments.items():
            annotation = signature.parameters[name].annotation
            if annotation not in (None, inspect._empty) and not isinstance(value, annotation):
                raise TypeError(
                    f"parameter ({name}) supposed to be of type ({annotation.__name__}), but found ({type(value).__name__})"
                )

        # call method
        result = func(*bound.args, **bound.kwargs)
        # verify result type
        return_type = signature.return_annotation
        if return_type is inspect._empty or return_type is None:
            return_type = type(None)

        if not isinstance(result, return_type):
            raise TypeError(
                f"method is supposed to return ({return_type}), but it returned ({type(result)})"
            )

        return result
    return wrapper


class BaseActor:
    def __init__(self):
        self.path = None

    @actor_method
    def info(self) -> dict:
        info = {}
        info["path"] = self.path
        info["methods"] = {}

        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        for name, attr in methods:
            if name.startswith("_"):
                continue

            signature = inspect.signature(attr)
            info["methods"][name] = {}
            info["methods"][name]["args"] = []
            info["methods"][name]["doc"] = attr.__doc__ or ""

            for parameter_name, parameter in signature.parameters.items():
                info["methods"][name]["args"].append((parameter_name, parameter.annotation.__name__))

        return info

    def __validate_actor__(self):
        def validate_annotation(annotation, annotated):
            if annotation is None or annotation is inspect._empty:
                return

            if not (inspect.isclass(annotation) and annotation.__class__ == type):
                raise ValueError("annotation must be a class type")

            if annotation not in (str, int, float, list, tuple, dict, bool):
                if annotation.__module__ == "builtins":
                    raise ValueError(f"unsupported type ({annotation.__name__})")

                for method in ["to_dict", "from_dict"]:
                    if method not in dir(annotation):
                        raise ValueError(
                            f"type ({annotation.__name__}) which annotate {annotated} doesn't have {method} method"
                        )

        result = {"valid": True, "errors": {}}
        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        for method_name, method_callable in methods:
            if method_name.startswith("_"):
                continue

            result["errors"][method_name] = []
            signature = inspect.signature(method_callable)
            try:
                validate_annotation(signature.return_annotation, "return")
            except ValueError as e:
                result["errors"][method_name].append(str(e))

            for name, parameter in signature.parameters.items():
                try:
                    validate_annotation(parameter.annotation, f"parameter ({name})")
                except ValueError as e:
                    result["errors"][method_name].append(str(e))

        if any(result["errors"].values()):
            result["valid"] = False

        return result



"""Loads a module from path

Let's make sure we have a python module in path `/tmp/hello.py`
```
~> cat /tmp/hello.py
a = 5
b = 6
z = a+b
```

## Using the codeloader tool

```python
JS-NG> m = j.tools.codeloader.load_python_module("/tmp/hello.py")
JS-NG> m.a
5
```

"""

import os
import sys
import importlib


def load_python_module(module_path: str):
    """Loads python module by module path

    Arguments:
        module_path {str} -- absolute path of the module
    """
    from jumpscale.loader import j

    module_uid = module_path[:-3]
    module_name = j.sals.fs.stem(module_path)
    spec = importlib.util.spec_from_file_location(module_name, module_path)

    if module_uid in sys.modules:
        return sys.modules[module_uid]

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_uid] = module
    spec.loader.exec_module(module)

    return module



"""Generator for crystal language. Takes in parsed schemas and generates crystal language classes.
"""
from .plugin import Plugin
from jumpscale.loader import j

types_map = {
    "": "String",
    "S": "String",
    "O": "Object",
    "I": "Int64",
    "F": "Float",
    "B": "Boolean",
    "L": "[] of Object",
    "LS": "[] of String",
    "LI": "[] of Int64",
    "LF": "[] of Float",
    "LO": "",
}


def get_prop_line(prop):
    prop_type = prop.prop_type
    crystal_type = types_map.get(prop_type)
    line = f"property {prop.name}"

    # print(f"\n\n{prop.name} => {prop} \n\n")
    # primitive with a default or not.
    if prop_type == "E":
        line += f" : {prop.name.capitalize()}"
    elif prop_type == "O":
        line += f" : {prop.url_to_class_name}"
    elif prop_type == "LO" and prop.defaultvalue and prop.defaultvalue != "[]":
        line += f" : [] of {prop.url_to_class_name}"
    elif prop_type == "LO" and not prop.defaultvalue:
        line += f" : [] of Object"
    elif crystal_type == "L" and not prop.defaultvalue:
        line += f" : [] of Object"
    elif crystal_type == "L" and prop.defaultvalue:
        line += f" = {crystal_type}"
    elif len(prop_type) > 1 and prop_type[0] == "L" and prop_type[1] != "O":
        line += f" : {crystal_type}"

    elif prop_type in ["I", "F", "B"] and not prop.defaultvalue:
        line += f" : {crystal_type}"
    elif prop_type in ["I", "F", "B"] and prop.defaultvalue:
        line += f" = {prop.defaultvalue}"
    elif prop_type == "S":
        line += f' = "{prop.defaultvalue}"'
    else:
        line += f": {crystal_type}"

    return line


SINGLE_TEMPLATE = """

class {{generated_class_name}}
{%- for prop in generated_properties.values() %}
    {{get_prop_line(prop)}}
{%- endfor %}
end

"""

TEMPLATE = """
#GENERATED CLASS DONT EDIT

{%- for enum in enums %}

enum {{enum['name']}}:
    {%- for enumval in enum['vals'] %}
    {{enumval}} = {{loop.index0}}
    {%- endfor %}
end
{%- endfor %}


{{classes_generated}}

"""


class CrystalGenerator(Plugin):
    def __init__(self):
        super().__init__()

        self._single_template = SINGLE_TEMPLATE
        self._template = TEMPLATE
        self._get_prop_line = get_prop_line
        self._types_map = types_map



"""
Executor Local allows executing commands within specific env on the local machine. using the executor framework you can retrieve the stdout, stderr, and the return code as well.
```
JS-NG> j.core.executors.run_local("echo $WHOME", env={'WHOME':'abc'}, ech
     1 o=True)
echo $WHOME
abc
<Result cmd='echo $WHOME' exited=0>

JS-NG> j.core.executors.run_local("uname", echo=True)
uname
Linux
<Result cmd='uname' exited=0>

JS-NG> j.core.executors.run_local("uname", echo=False)
Linux
<Result cmd='uname' exited=0>
```


"""
import invoke

from .command_builder import cmd_from_args


@cmd_from_args
def execute(cmd, **command_ctx):
    """
    :param str shell:
        Which shell binary to use. Default: ``/bin/bash`` (on Unix;
        ``COMSPEC`` or ``cmd.exe`` on Windows.)
    :param bool warn:
        Whether to warn and continue, instead of raising
        `.UnexpectedExit`, when the executed command exits with a
        nonzero status. Default: ``False``.
        .. note::
            This setting has no effect on exceptions, which will still be
            raised, typically bundled in `.ThreadException` objects if they
            were raised by the IO worker threads.
            Similarly, `.WatcherError` exceptions raised by
            `.StreamWatcher` instances will also ignore this setting, and
            will usually be bundled inside `.Failure` objects (in order to
            preserve the execution context).
            Ditto `.CommandTimedOut` - basically, anything that prevents a
            command from actually getting to "exited with an exit code"
            ignores this flag.
    :param hide:
        Allows the caller to disable ``run``'s default behavior of copying
        the subprocess' stdout and stderr to the controlling terminal.
        Specify ``hide='out'`` (or ``'stdout'``) to hide only the stdout
        stream, ``hide='err'`` (or ``'stderr'``) to hide only stderr, or
        ``hide='both'`` (or ``True``) to hide both streams.
        The default value is ``None``, meaning to print everything;
        ``False`` will also disable hiding.
        .. note::
            Stdout and stderr are always captured and stored in the
            ``Result`` object, regardless of ``hide``'s value.
        .. note::
            ``hide=True`` will also override ``echo=True`` if both are
            given (either as kwargs or via config/CLI).
    :param bool pty:
        By default, ``run`` connects directly to the invoked process and
        reads its stdout/stderr streams. Some programs will buffer (or even
        behave) differently in this situation compared to using an actual
        terminal or pseudoterminal (pty). To use a pty instead of the
        default behavior, specify ``pty=True``.
        .. warning::
            Due to their nature, ptys have a single output stream, so the
            ability to tell stdout apart from stderr is **not possible**
            when ``pty=True``. As such, all output will appear on
            ``out_stream`` (see below) and be captured into the ``stdout``
            result attribute. ``err_stream`` and ``stderr`` will always be
            empty when ``pty=True``.
    :param bool fallback:
        Controls auto-fallback behavior re: problems offering a pty when
        ``pty=True``. Whether this has any effect depends on the specific
        `Runner` subclass being invoked. Default: ``True``.
    :param bool echo:
        Controls whether `.run` prints the command string to local stdout
        prior to executing it. Default: ``False``.
        .. note::
            ``hide=True`` will override ``echo=True`` if both are given.
    :param dict env:
        By default, subprocesses receive a copy of Invoke's own environment
        (i.e. ``os.environ``). Supply a dict here to update that child
        environment.
        For example, ``run('command', env={'PYTHONPATH':
        '/some/virtual/env/maybe'})`` would modify the ``PYTHONPATH`` env
        var, with the rest of the child's env looking identical to the
        parent.
        .. seealso:: ``replace_env`` for changing 'update' to 'replace'.
    :param bool replace_env:
        When ``True``, causes the subprocess to receive the dictionary
        given to ``env`` as its entire shell environment, instead of
        updating a copy of ``os.environ`` (which is the default behavior).
        Default: ``False``.
    :param str encoding:
        Override auto-detection of which encoding the subprocess is using
        for its stdout/stderr streams (which defaults to the return value
        of `default_encoding`).
    :param out_stream:
        A file-like stream object to which the subprocess' standard output
        should be written. If ``None`` (the default), ``sys.stdout`` will
        be used.
    :param err_stream:
        Same as ``out_stream``, except for standard error, and defaulting
        to ``sys.stderr``.
    :param in_stream:
        A file-like stream object to used as the subprocess' standard
        input. If ``None`` (the default), ``sys.stdin`` will be used.
        If ``False``, will disable stdin mirroring entirely (though other
        functionality which writes to the subprocess' stdin, such as
        autoresponding, will still function.) Disabling stdin mirroring can
        help when ``sys.stdin`` is a misbehaving non-stream object, such as
        under test harnesses or headless command runners.
    :param watchers:
        A list of `.StreamWatcher` instances which will be used to scan the
        program's ``stdout`` or ``stderr`` and may write into its ``stdin``
        (typically ``str`` or ``bytes`` objects depending on Python
        version) in response to patterns or other heuristics.
        See :doc:`/concepts/watchers` for details on this functionality.
        Default: ``[]``.
    :param bool echo_stdin:
        Whether to write data from ``in_stream`` back to ``out_stream``.
        In other words, in normal interactive usage, this parameter
        controls whether Invoke mirrors what you type back to your
        terminal.
        By default (when ``None``), this behavior is triggered by the
        following:
            * Not using a pty to run the subcommand (i.e. ``pty=False``),
                as ptys natively echo stdin to stdout on their own;
            * And when the controlling terminal of Invoke itself (as per
                ``in_stream``) appears to be a valid terminal device or TTY.
                (Specifically, when `~invoke.util.isatty` yields a ``True``
                result when given ``in_stream``.)
                .. note::
                    This property tends to be ``False`` when piping another
                    program's output into an Invoke session, or when running
                    Invoke within another program (e.g. running Invoke from
                    itself).
        If both of those properties are true, echoing will occur; if either
        is false, no echoing will be performed.
        When not ``None``, this parameter will override that auto-detection
        and force, or disable, echoing.
    :param timeout:
    """
    ## use formatter to format command
    command_ctx = command_ctx or {}
    if "cwd" in command_ctx:
        if command_ctx["cwd"]:
            cwd = command_ctx["cwd"]
            del command_ctx["cwd"]
            c = invoke.Context()
            with c.cd(cwd):
                res = c.run(cmd, **command_ctx)
                return res.return_code, res.stdout, res.stderr
        del command_ctx["cwd"]
    res = invoke.run(cmd, **command_ctx)
    return res.return_code, res.stdout, res.stderr



from .threebot import ThreeBot as threebot
import inspect
import cgi
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import print_formatted_text

__all__ = ["threebot", "info"]


def info():
    print_formatted_text(HTML(_get_doc(__all__)))


def _get_doc_line(doc):
    if not doc:
        return ""
    for line in doc.splitlines():
        if line.strip():
            return line.strip()
    return ""


def _get_doc(root_module, level=0, size=4):
    """get a formatted docstring from a module
    this will loop over __all__self.

    :param root_module: root module
    :type root_module: module
    :param level: spacing level, defaults to 0
    :type level: int, optional
    :param level: spacing size, defaults to 4
    :type level: int, optional
    :return: docstring
    :rtype: str
    """

    doc = ""

    if isinstance(root_module, list):
        glob = globals()
        members = [(name, glob[name]) for name in root_module]
    else:
        members = inspect.getmembers(root_module)
    for name, obj in members:
        if name.startswith("_"):
            continue
        if name[0].lower() != name[0]:
            continue

        is_module = not inspect.isroutine(obj)
        if is_module and level != 0:
            continue

        spaces = " " * level

        if is_module:
            doc += f"{spaces}<ansibrightblue>{name}</ansibrightblue>"
        elif getattr(obj, "__property__", False):
            doc += f"{spaces}<ansicyan>{name}</ansicyan>"
        else:
            doc += f"{spaces}<ansigreen>{name}</ansigreen>"

        if obj.__doc__:
            try:
                # only get first line of member docstring
                first_line = _get_doc_line(obj.__doc__)
                doc += cgi.html.escape(f": {first_line}")
            except IndexError:
                pass

        doc = f"{doc}\n"

        if is_module:
            doc += _get_doc(obj, level=level + size)

    return doc



def main():
    import sqlite3
    from pprint import pprint

    conn = sqlite3.connect("test_index.db")
    c = conn.cursor()
    c.execute("SELECT * FROM employee_name")
    pprint(c.fetchall())
    c.execute("SELECT * FROM employee_salary")
    pprint(c.fetchall())


if __name__ == "__main__":
    main()



import redis

from . import ConfigNotFound, EncryptedConfigStore
from .serializers import JsonSerializer


class RedisStore(EncryptedConfigStore):
    """
    RedisStore store is an EncryptedConfigStore

    It saves the data in redis and configuration for redis comes from `config_env.get_store_config("redis")`
    """

    def __init__(self, location):
        """
        create a new redis store, the location given will be used to generate keys

        this keys will be combined to get/set instance config

        Args:
            location (Location)
        """
        super().__init__(location, JsonSerializer())
        redis_config = self.config_env.get_store_config("redis")
        self.redis_client = redis.Redis(redis_config["hostname"], redis_config["port"])

    def get_key(self, instance_name):
        """
        get a key for an instance

        this will return a dot-separated key derived from current location

        Args:
            instance_name (str): name

        Returns:
            str: key
        """
        return ".".join([self.location.name, instance_name])

    def read(self, instance_name):
        """
        read instance config from redis

        Args:
            instance_name (name): name

        Returns:
            str: data
        """
        key = self.get_key(instance_name)
        if not self.redis_client.exists(key):
            raise ConfigNotFound(f"cannot find config for {instance_name} at {key}")
        return self.redis_client.get(key)

    def _full_scan(self, pattern):
        """
        get the full result of a scan command on current redis database by this pattern

        Args:
            pattern ([type]): [description]
        """
        keys = []
        cursor, values = self.redis_client.scan(0, pattern)

        while values:
            keys += values
            if not cursor:
                break
            cursor, values = self.redis_client.scan(cursor, pattern)

        return keys

    def get_location_keys(self):
        """
        get all keys under current location (scanned)

        Returns:
            list: a list of keys
        """
        return self._full_scan(f"{self.location.name}.*")

    def list_all(self):
        """
        get all names of instances (instance keys)

        Returns:
            [type]: [description]
        """
        names = []

        keys = self.get_location_keys()
        for key in keys:
            # remove location name part
            name = key.decode().replace(self.location.name, "").lstrip(".")
            # if it does not contain a ".", then it's not a child
            if "." not in name:
                names.append(name)
        return names

    def write(self, instance_name, data):
        """
        set data with the corresponding key for this instance

        Args:
            instance_name (str): name
            data (str): data

        Returns:
            bool: written or not
        """
        return self.redis_client.set(self.get_key(instance_name), data)

    def delete(self, instance_name):
        """
        delete given instance

        Args:
            instance_name (str): name

        Returns:
            bool
        """
        return self.redis_client.delete(self.get_key(instance_name))



"""This module used to manage your conatiners, create ,run, list, delete and exec commands on docker
for example
```
#getting docker clinet
cl = j.clients.docker.get('test')
running docker
cl.run("container_name", "threefoldtech/js-ng")
# listing contaienrs
cl.list(all=True) # lists all containers include stopped ones
# getting container
container = cl.get(<container_id or name>)
executing commmands
cl.exec("container_id", "ls /tmp")
or container.exec_run("ls /tmp)
```
"""
import docker
from jumpscale.clients.base import Client
from jumpscale.core.base import fields


class DockerClient(Client):
    base_url = fields.String()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__client = None

    @property
    def client(self):
        if not self.__client:
            if self.base_url:
                self.__client = docker.DockerClient(base_url=self.base_url)
            else:
                self.__client = docker.from_env()
        return self.__client

    def create(
        self,
        name,
        image,
        command=None,
        environment=None,
        entrypoint=None,
        volumes=None,
        devices=None,
        detach=True,
        ports=None,
        privileged=False,
        auto_remove=False,
        hostname="js-ng",
    ):
        """Creates a docker container without starting it

        Args:
            name (str): name of the docker container
            image (str): image of the docker container
            command (str or list) – The command to run in the container.
            environment (dict or list) – Environment variables to set inside the container, as a dictionary or a list of strings in the format ["SOMEVARIABLE=xxx"].
            entrypoint (str or list) – The entrypoint for the container.
            volumes (dict or list) –
                A dictionary to configure volumes mounted inside the container. The key is either the host path or a volume name, and the value is a dictionary with the keys:
                bind The path to mount the volume inside the container
                mode Either rw to mount the volume read/write, or ro to mount it read-only.
                example
                {'/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
                '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}}
            devices (list) –
                Expose host devices to the container, as a list of strings in the form <path_on_host>:<path_in_container>:<cgroup_permissions>.
                For example, /dev/sda:/dev/xvda:rwm allows the container to have read-write access to the host’s /dev/sda via a node named /dev/xvda inside the container.
            detach (bool, optional): detach from container after running it.
            ports The port number, as an integer. For example,
                - {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
                - None, to assign a random host port. For example, {'2222/tcp': None}.
                - A tuple of (address, port) if you want to specify the host interface. For example, {'1111/tcp': ('127.0.0.1', 1111)}.
                - A list of integers, if you want to bind multiple host ports to a single container port. For example, {'1111/tcp': [1234, 4567]}.
            privileged (bool) – Give extended privileges to this container.
            auto_remove (bool) – enable auto-removal of the container on daemon side
            hostname (str) - hostname to be set on docker container default "js-ng"
        Returns:
            container: container object
        """
        docker = self.client.containers.create(
            name=name,
            image=image,
            command=command,
            environment=environment,
            entrypoint=entrypoint,
            volumes=volumes,
            devices=devices,
            detach=detach,
            ports=ports,
            privileged=privileged,
            auto_remove=auto_remove,
            hostname=hostname,
        )
        return docker

    def run(
        self,
        name,
        image,
        command=None,
        environment=None,
        entrypoint=None,
        volumes=None,
        devices=None,
        detach=True,
        ports=None,
        privileged=False,
        auto_remove=False,
        hostname="js-ng",
    ):
        """Runs docker container
        Args:
            name (str): name of the docker container
            image (str): image of the docker container
            command (str or list) – The command to run in the container.
            environment (dict or list) – Environment variables to set inside the container, as a dictionary or a list of strings in the format ["SOMEVARIABLE=xxx"].
            entrypoint (str or list) – The entrypoint for the container.
            volumes (dict or list) –
                A dictionary to configure volumes mounted inside the container. The key is either the host path or a volume name, and the value is a dictionary with the keys:
                bind The path to mount the volume inside the container
                mode Either rw to mount the volume read/write, or ro to mount it read-only.
                example
                {'/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
                '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}}
            devices (list) –
                Expose host devices to the container, as a list of strings in the form <path_on_host>:<path_in_container>:<cgroup_permissions>.
                For example, /dev/sda:/dev/xvda:rwm allows the container to have read-write access to the host’s /dev/sda via a node named /dev/xvda inside the container.
            detach (bool, optional): detach from container after running it.
            ports The port number, as an integer. For example,
                - {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
                - None, to assign a random host port. For example, {'2222/tcp': None}.
                - A tuple of (address, port) if you want to specify the host interface. For example, {'1111/tcp': ('127.0.0.1', 1111)}.
                - A list of integers, if you want to bind multiple host ports to a single container port. For example, {'1111/tcp': [1234, 4567]}.
            privileged (bool) – Give extended privileges to this container.
            auto_remove (bool) – enable auto-removal of the container on daemon side
            hostname (str) - hostname to be set on docker container default "js-ng"
        Returns:
            container: container object
        """
        docker = self.client.containers.run(
            name=name,
            image=image,
            command=command,
            environment=environment,
            entrypoint=entrypoint,
            volumes=volumes,
            devices=devices,
            detach=detach,
            ports=ports,
            privileged=privileged,
            auto_remove=auto_remove,
            hostname=hostname,
        )
        return docker

    def get(self, container_id):
        """Runs docker container
        Args:
            container_id (str): Id or name of the docker container

        Returns:
            container: container object
        """
        container = self.client.containers.get(container_id)
        return container

    def exists(self, container_id):
        try:
            self.client.containers.get(container_id)
            return True
        except docker.errors.NotFound:
            return False

    def delete(self, container_id, force=False):
        """Deletes docker container
        Args:
            container_id (str): Id or name of the docker container
            force (bool): image of the docker container

        Returns:
            bool: True container deleted
        """
        container = self.get(container_id)
        container.remove(force=force)
        return True

    def kill(self, container_id, sig=None):
        """Kills docker container
        Args:
            container_id (str): Id or name of the docker container
            sig (str or int): kill siganle, default SIGKILL

        Returns:
            bool: True container killed
        """
        container = self.get(container_id)
        container.kill(sig)
        return True

    def restart(self, container_id, timeout=10):
        """Kills docker container
        Args:
            container_id (str): Id or name of the docker container
            timeout (int): timeout to stop container before trying to kill

        Returns:
            bool: True if container restarted
        """
        container = self.get(container_id)
        container.restart(timeout=timeout)
        return True

    def start(self, container_id):
        """starts docker container
        Args:
            container_id (str): Id or name of the docker container
            timeout (int): timeout to stop container before trying to kill

        Returns:
            bool: True if container started
        """
        container = self.get(container_id)
        container.start()
        return True

    def stop(self, container_id):
        """stops docker container
        Args:
            container_id (str): Id or name of the docker container
            timeout (int): timeout to stop container before trying to kill

        Returns:
            bool: True if container started
        """
        container = self.get(container_id)
        container.stop()
        return True

    def list(self, all=False):
        """Returns list of docker containers created
        Args:
            all (bool): list all containers including stopped ones

        Returns:
            list: list of containers
        """
        return self.client.containers.list(all=all)

    def exec(
        self,
        container_id,
        cmd,
        stdout=True,
        stderr=True,
        stdin=False,
        tty=False,
        privileged=False,
        user="",
        detach=False,
        stream=False,
        socket=False,
        environment=None,
        workdir=None,
        demux=True,
    ):
        """Executes command docker container
        Args:
            container_id (str): Id or name of the docker container
            cmd (str or list) – Command to be executed
            stdout (bool) – Attach to stdout. Default: True
            stderr (bool) – Attach to stderr. Default: True
            stdin (bool) – Attach to stdin. Default: False
            tty (bool) – Allocate a pseudo-TTY. Default: False
            privileged (bool) – Run as privileged.
            user (str) – User to execute command as. Default: root
            detach (bool) – If true, detach from the exec command. Default: False
            stream (bool) – Stream response data. Default: False
            socket (bool) – Return the connection socket to allow custom read/write operations. Default: False
            environment (dict or list) – A dictionary or a list of strings in the following format ["PASSWORD=xxx"] or {"PASSWORD": "xxx"}.
            workdir (str) – Path to working directory for this exec session
            demux (bool) – Return stdout and stderr separately

        Returns:
            A tuple of (exit_code, output)
            exit_code: (int):
            Exit code for the executed command or None if either stream or socket is True.

            output: (generator, bytes, or tuple):
                If stream=True, a generator yielding response chunks.
                If socket=True, a socket object for the connection.
                If demux=True, a tuple of two bytes: stdout and stderr.
                A bytestring containing response data otherwise.
        """
        docker = self.get(container_id)
        return docker.exec_run(
            cmd=cmd,
            stdout=stdout,
            stderr=stderr,
            stdin=stdin,
            tty=tty,
            privileged=privileged,
            user=user,
            detach=detach,
            stream=stream,
            socket=socket,
            environment=environment,
            workdir=workdir,
            demux=demux,
        )



"""

SSHClient modules helps connecting and executing commands to a remote machine.


## Create SSH Key
```python
JS-NG> xmonader = j.clients.sshkey.new("xmonader")
JS-NG> xmonader.private_key_path = "/home/xmonader/.ssh/id_rsa"
JS-NG>

```

## Creating sshclient using that key and executing commands

```
JS-NG> sshkey = j.clients.sshkey.get("xmonader")
JS-NG> localclient = j.clients.sshclient.get("xmonader")
JS-NG> localclient.host = "IP of the machine to access"
JS-NG> localclient.sshclient.run("hostname")
asgard
(0, 'asgard\n', '')

```
"""

from jumpscale.clients.base import Client
from jumpscale.core.base import fields
from jumpscale.loader import j


class SSHClient(Client):
    """
    SSHClient has the following properties:
    sshkey (str): sshkey to use within that client
    host (str): host ip
    user (str): user to connect as default: True
    port (int): the port to use
    forward_agent (bool):  forward agent or not (default True)
    connect_timeout (int): timeout (default 10 seconds)

    """

    sshkey = fields.String(required=True)

    host = fields.String(default="127.0.0.1", required=True)
    user = fields.String(default="root", required=True)
    port = fields.Integer(default=22, required=True)
    forward_agent = fields.Boolean(default=True)
    connect_timeout = fields.Integer(default=10)
    connection_kwargs = fields.Typed(dict, default={})

    # gateway = ?  FIXME: should help with proxyjumps. http://docs.fabfile.org/en/2.4/concepts/networking.html#ssh-gateways

    inline_ssh_env = fields.Boolean(
        default=True
    )  # whether to send environment variables “inline” as prefixes in front of command strings (export VARNAME=value && mycommand here), instead of trying to submit them through the SSH protocol itself (which is the default behavior). This is necessary if the remote server has a restricted AcceptEnv setting (which is the common default).

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__client = None

    @property
    def _sshkey(self):
        """ Get sshkey client that you have loaded
        e.g
            JS-NG> localconnection = j.clients.sshclient.new("localconnection")
            JS-NG> localconnection.sshkey = "xmonader"
            JS-NG> localconnection._sshkey()  -> SHKeyClient(_Base__instance_name='xmonader', _Base__parent=None, ...
        Returns:
            Obj: It returns object of SSHkeyClient
        """
        return j.clients.sshkey.get(self.sshkey)

    @property
    def sshclient(self):
        self.validate()
        if not self.__client:
            self.connection_kwargs["key_filename"] = self._sshkey.private_key_path
            connection_kwargs = dict(
                host=self.host,
                user=self.user,
                port=self.port,
                forward_agent=self.forward_agent,
                connect_timeout=self.connect_timeout,
                connect_kwargs=self.connection_kwargs,
            )
            if self._sshkey.passphrase:
                connection_kwargs["connect_kwargs"]["passphrase"] = self._sshkey.passphrase

            self.__client = j.core.executors.RemoteExecutor(**connection_kwargs)

        return self.__client

    def reset_connection(self):
        """ Reset the connection
        e.g
            localconnection = j.clients.sshclient.new("localconnection")
            localconnection.reset_connection()

        """
        self.__client = None



import ast


class String:
    def __init__(self, default=None):
        if not default:
            default = ""
        self.default = default

    def check(self, value):
        """Check whether provided value is a string"""
        return isinstance(value, str)

    def from_str(self, value):
        """return string from a string (is basically no more than a check)"""
        return value


class Integer:
    def __init__(self, default=None):
        if not default:
            default = 0
        self.default = self.from_str(default)

    def check(self, value):
        """Check whether provided string represent integer value
        
        Arguments:
            value (str)
        """
        try:
            int(value)
            return True
        except ValueError:
            return False

    def from_str(self, value):
        """get integer value from tha string
        
        Arguments:
            value (str)
        """
        try:
            return int(value)
        except ValueError:
            return None


class Boolean:
    def __init__(self, default=None):
        if not default:
            default = "False"
        self.default = self.from_str(default)

    def check(self, value):
        """Check whether provided string represent Boolean expresion

        Arguments:
            value (str)
        """
        valid_expresions = ["True", "False"]
        return value in valid_expresions

    def from_str(self, value):
        """Get the Boolean value from the string

        Arguments:
            value (str)
        """
        if value == "True":
            return True
        elif value == "False":
            return False
        else:
            return None


class Float:
    def __init__(self, default=None):
        if not default:
            default = 0.0
        self.default = self.from_str(default)

    def check(self, value):
        """Check whether provided string represent integer value
        
        Arguments:
            value (str)
        """
        try:
            float(value)
            return True
        except ValueError:
            return False

    def from_str(self, value):
        """get integer value from tha string
        
        Arguments:
            value (str)
        """
        try:
            return float(value)
        except ValueError:
            return None


class JSObject:
    def __init__(self, default=None):
        default = default or ""
        self.default = self.from_str(default)

    def check(self, value):
        """Check whether provided string represent JSObject value. (Any string will do).
        
        Arguments:
            value (str)
        """
        return isinstance(value, str)

    def from_str(self, value):
        """Return value as is.
        
        Arguments:
            value (str)
        """
        return value


class List:
    def __init__(self, default, subtype):
        self.subtype = subtype
        default = default or "[]"
        self.default = self.from_str(default)

    def _deep_check(self, value):
        """Check that the value represents a list with proper elements of the specified subtype.
        
        Args:
            value (list): The list to be checked.
        
        Returns:
            Boolean: True if the list is valid.
        """
        if not isinstance(value, list):
            return False
        for e in value:
            if isinstance(self.subtype, List) and not self.subtype._deep_check(e):
                return False
            elif not isinstance(self.subtype, List) and not self.subtype.check(str(e)):
                return False
        return True

    def check(self, value):
        """Check that the value represents a list with proper elements of the specified subtype.
        
        Args:
            value (list): The list to be checked.
        
        Returns:
            Boolean: True if the list is valid.
        """
        valid = len(value) >= 2 and value[0] == "[" and value[-1] == "]"
        if not valid:
            return False
        try:
            return self._deep_check(ast.literal_eval(value))
        except:
            return False
        return True

    def _deep_parse(self, value):
        """parses the subelements (if they are of different python type it's converted using the subtype parser)
        
        Args:
            value (list): The list to be parsed.
        
        Returns:
            list: The parsed list.
        """
        result = []
        if isinstance(self.subtype, List):
            for i, e in enumerate(value):
                value[i] = self.subtype._deep_parse(e)
        else:
            for i, e in enumerate(value):
                value[i] = self.subtype.from_str(str(e))
        return value

    def from_str(self, value):
        """parses the string value into a list.
        
        Args:
            value (str): The string to be parsed.
        
        Returns:
            list: The parsed list.
        """
        if isinstance(self.subtype, JSObject):
            return value
        else:
            return self._deep_parse(ast.literal_eval(value))


def get_js_type(type_str, default_value=None):
    """Gets the js type from a string.

    1. "S" -> String
    2. "B" -> Boolean
    3. "I" -> Integer
    4. "O" -> JSObject
    4. "F" -> Float
    5. "L.*" -> List with subtype .*
    6. "" -> empty defaults to String
    
    Args:
        type_str (str): type description.
        default_value (any, optional): The default value. Defaults to None.
    
    Returns:
        Object: A js type object.
    """
    types = {"": String, "S": String, "I": Integer, "B": Boolean, "F": Float, "O": JSObject}
    if len(type_str) == 0 or type_str[0] != "L":
        return types[type_str](default_value)
    else:
        subtype = type_str[1:]
        return List(default_value, get_js_type(subtype))



import yaml


def dumps(obj):
    """dump dict object into yaml stream 
    
    Arguments:
        obj (dict) : the dict which will be dumped     
    
    Returns:
        string : the yaml stream
    """
    return yaml.dump(obj)


def loads(s):
    """loads the data from yaml string into dict
    
    Arguments:
        s (string) : the yaml stream
    
    Returns:
        dict : the loaded data from yaml stram
    """
    return yaml.load(s)




import re


class HostsFile:
    def __init__(self, hosts_file_path):
        self.path = hosts_file_path
        self.content = self._parse()

    def _parse(self):
        with open(self.path, "r") as file:
            content_text = file.read()
        content_lines = content_text.splitlines()
        ip_lines = []
        regex = r"""\d+\.\d+\.\d+\.\d+"""
        for line in content_lines:
            if re.search(regex, line):
                ip_lines.append(line)
        content_dict = {}
        for line in ip_lines:
            key, value = line.split("\t")[0], line.split("\t")[1]
            content_dict[key] = value
        return content_dict

    def write(self):
        """
        write the changes into the file.
        """
        content_text = ""
        for key, value in self.content.items():
            content_text += f"\n{key}\t{value}"
        with open(self.path, "w") as file:
            file.write(content_text)

    def remove(self, ip):
        """
        remove the ip and its hostname from hosts file
        Args:
            ip (str) : the ip address
        """
        del self.content[ip]

    def add(self, ip, domain):
        """
        add new entry to the hosts file
        Args:
            ip (str) : the ip address
            domain (str) : the host name
        """
        self.content[ip] = domain

    def set_hostname(self, ip, domain):
        """
        update the hostname for ip
        Args:
            ip (str) : the ip address
            domain (str) : the host name           
        """
        self.content[ip] = domain

    def exists(self, ip):
        """
        check for the existence of the ip in hosts file.
        Args:
            ip (str) : the ip address
        Return:
            boolen expression
        """
        return ip in self.content

    def get_hostname(self, ip):
        """
        get the hostname for ip
        Args:
            ip (str) : the ip address
        Returns:
            the hostname for the ip address
        """
        return self.content[ip]



"""
Base and factory related events.
"""


class InstanceEvent:
    def __init__(self, instance=None, factory=None):
        self.instance = instance
        self.factory = factory


class AttributeUpdateEvent(InstanceEvent):
    def __init__(self, instance, name, new_value):
        super().__init__(instance=instance)
        self.name = name
        self.new_value = new_value


class InstanceCreateEvent(InstanceEvent):
    pass


class InstanceDeleteEvent(InstanceEvent):
    def __init__(self, name, instance=None, factory=None):
        super().__init__(instance, factory)
        self.name = name



"""
Executor remote allows executing commands within specific env on the local machine. using the executor framework you can retrieve the stdout, stderr, and the return code as well.

```
JS-NG> with j.core.executors.RemoteExecutor(host="localhost", connect_kwargs={"key_filename":
                                                                              "/home/xmonader/.ssh/id_rsa",}) as c:
            c.run("hostname")
xmonader-ThinkPad-E580
JS-NG>
```
"""


import fabric

from .command_builder import cmd_from_args


@cmd_from_args
def execute(cmd, command_ctx, connection_ctx):
    """
    execute a command on a remote context

    Args:
        cmd (str or list): command as a string or an argument list, e.g. `"ls -la"` or `["ls", "la"]`
        command_ctx (dict): command runner context (the same as local `execute`)
        connection_ctx (dict): context passed to fabric e.g. fabric.Connection(host, user=None, port=None, config=None, gateway=None, forward_agent=None, connect_timeout=None, connect_kwargs=None, inline_ssh_env=None)

    Returns:
        tuple: return code, stdout, stderr
    """
    with fabric.Connection(**connection_ctx) as c:
        res = c.run(cmd, **command_ctx)
        return res.return_code, res.stdout, res.stderr


class RemoteExecutor:
    def __init__(self, **connection_ctx):
        self._connection_ctx = connection_ctx

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        pass

    @property
    def connection(self):
        return fabric.Connection(**self._connection_ctx)

    @property
    def sftp(self):
        return self.connection.sftp()

    def run(self, cmd, **command_ctx):
        """
        execute a command

        Args:
            cmd (str or list): "ls -la" or ["ls", "-la"]

        Returns:
            tuple: return code, stdout, stderr
        """
        return execute(cmd, command_ctx, self._connection_ctx)



from jumpscale.clients.docker.docker import DockerClient
from jumpscale.core.dirs.dirs import Dirs
from jumpscale.core.exceptions import Value
from jumpscale.core.executors.local import execute

docker_client = DockerClient()


class Container:
    """Container management
    """

    @staticmethod
    def install(name, image, development: bool = False, volumes=None):
        """Creates a container

        Args:
            name (str): name of the container
            image (str): container image.
            development (bool, optional): if true will mount codedir. Defaults to False.
            volumes (dict, optional): paths to be mounted

        Raises:
            Value: Container with specified name already exists
        """
        if docker_client.exists(name):
            raise Value("Container with specified name already exists")

        volumes = volumes or {}
        if development:
            volumes = {Dirs.CODEDIR: {"bind": "/sandbox/code", "mode": "rw"}}

        print(f"Creating container {name}")
        return docker_client.run(name, image, entrypoint="/sbin/my_init", volumes=volumes, detach=True)

    @staticmethod
    def start(name):
        """Starts an existing container

        Args:
            name (str): name of the container
        """
        if not docker_client.exists(name):
            raise Value("Container with specified name doesn't exist")
        docker_client.start(name)

    @staticmethod
    def exec(name, cmd):
        """Execute command in container

        Args:
            name (str): name of the container
            cmd (str or list): command to execute
        """
        basecmd = ["docker", "exec", "-it", name]
        if isinstance(cmd, str):
            basecmd.append(cmd)
        else:
            basecmd += cmd
        execute(basecmd, pty=True)

    @staticmethod
    def stop(name):
        """Stops an existing container

        Args:
            name (str): name of the container
        """
        if not docker_client.exists(name):
            raise Value("Container with specified name doesn't exist")
        docker_client.stop(name)

    @staticmethod
    def delete(name):
        Container.stop(name)
        docker_client.delete(name)



import os
import hashlib
import binascii
import itertools
import bisect

from .wordlist import WORDS as wordlist
from .exceptions import FailedChecksumError
from jumpscale.core import exceptions


def generate_mnemonic(strength=256, wordlist=wordlist):
    """Generate random mnemonic

    Args:
        strength (int, optional): The number of bits of the generated key. Defaults to 256.
        wordlist (list[str], optional): The Word list used to generate the mnemonic. Defaults to wordlist.

    Returns:
        str: A string of space separated words representing the mnemonic.
    """
    ent = entropy(strength)
    return key_to_mnemonic(ent)


def entropy(strength=256):
    """Returns random bytes of length strengh/8.

    Args:
        strength (int, optional): The number of bits in the randomly created array. Defaults to 256.

    Returns:
        bytes: The generated bytes.
    """
    return os.urandom(strength // 8)


def mnemonic_to_key(mnemonic, wordlist=wordlist):
    """Convert mnemonic to key

    Args:
        mnemonic (str): A string of space separated words representing the mnemonic.
        wordlist (list[string], optional): The Word list used to generate the mnemonic. Defaults to wordlist.

    Raises:
        FailedChecksumError: Thrown when the checksum validation fails.

    Returns:
        bytes: The key
    """
    words = mnemonic.split(" ")
    concatLenBits = len(words) * 11
    concatBits = [False] * concatLenBits
    wordindex = 0
    use_binary_search = True
    for word in words:
        # Find the words index in the wordlist
        ndx = binary_search(wordlist, word) if use_binary_search else wordlist.index(word)
        if ndx < 0:
            raise exceptions.NotFound('Unable to find "%s" in word list.' % word)
        # Set the next 11 bits to the value of the index.
        for ii in range(11):
            concatBits[(wordindex * 11) + ii] = (ndx & (1 << (10 - ii))) != 0
        wordindex += 1
    checksumLengthBits = concatLenBits // 33
    entropyLengthBits = concatLenBits - checksumLengthBits
    # Extract original entropy as bytes.
    entropy = bytearray(entropyLengthBits // 8)
    for ii in range(len(entropy)):
        for jj in range(8):
            if concatBits[(ii * 8) + jj]:
                entropy[ii] |= 1 << (7 - jj)
    # Take the digest of the entropy.
    hashBytes = hashlib.sha256(entropy).digest()
    hashBits = list(itertools.chain.from_iterable(([c & (1 << (7 - i)) != 0 for i in range(8)] for c in hashBytes)))
    # Check all the checksum bits.
    for i in range(checksumLengthBits):
        if concatBits[entropyLengthBits + i] != hashBits[i]:
            raise exceptions.Value("Failed checksum.")
    return bytes(entropy)


def key_to_mnemonic(key, wordlist=wordlist):
    """Convert the passed key to memorizable mnemonic.

    Args:
        key (bytes): The key to be encoded.
        wordlist (list[str], optional): The wordlist used to generate the mnemonic. Defaults to wordlist.

    Returns:
        str: A string of space separated words representing the mnemonic
    """
    h = hashlib.sha256(key).hexdigest()
    b = (
        bin(int(binascii.hexlify(key), 16))[2:].zfill(len(key) * 8)
        + bin(int(h, 16))[2:].zfill(256)[: len(key) * 8 // 32]
    )
    result = []
    for i in range(len(b) // 11):
        idx = int(b[i * 11 : (i + 1) * 11], 2)
        result.append(wordlist[idx])
    result_phrase = " ".join(result)
    return result_phrase


def to_bin(arr, bytelen=8):
    """Converts an iterable object containing ascii codes to a string of the joint binary representation of each item where each of them is of length bytelen

    Args:
        arr (iterable): The array to be converted
        bytelen (int, optional): The length of the to-be-created binary string of each item. Defaults to 8.

    Returns:
        string: The binary string representing the array
    """
    result = ""
    for c in arr:
        result += ("0" * bytelen + bin(c)[2:])[-bytelen:]
    return result


def binary_search(a, x, lo=0, hi=None):  # can't use a to specify default for hi
    hi = hi if hi is not None else len(a)  # hi defaults to len(a)
    pos = bisect.bisect_left(a, x, lo, hi)  # find insertion position
    return pos if pos != hi and a[pos] == x else -1  # don't walk off the end
